"""
Simulates the dynamics of Figure 1A from
DOI: 10.1126/science.1101456.

Final output is a plot of the phenotypes
over time, based on sampling every 100
generations.
"""
import math
import sys

import attr
import matplotlib.pyplot as plt
import numpy as np

import fwdpy11


@attr.s()
class PySnowdrift(fwdpy11.PyDiploidGeneticValue):
    b1 = attr.ib()
    b2 = attr.ib()
    c1 = attr.ib()
    c2 = attr.ib()
    slope = attr.ib()  # "stretch" parameter for sigmoid function
    p0 = attr.ib()  # Phenotype of a mutant-free individual
    seed = attr.ib()

    def __attrs_post_init__(self):
        fwdpy11.PyDiploidGeneticValue.__init__(self, 1, None, None)
        self.phenotypes = []
        self.sig0 = (1.0 / self.slope) * math.log(self.p0 / (1.0 - self.p0))
        self.rng = fwdpy11.GSLrng(self.seed)

    def genetic_value_to_fitness(
        self, data: fwdpy11.DiploidGeneticValueToFitnessData
    ) -> float:
        zself = self.phenotypes[data.offspring_metadata.label]
        other = int(fwdpy11.gsl_ran_flat(self.rng, 0, len(self.phenotypes)))
        while other == data.offspring_metadata.label:
            other = int(fwdpy11.gsl_ran_flat(self.rng, 0, len(self.phenotypes)))
        zpair = zself + self.phenotypes[other]
        a = 1.0 + (
            self.b1 * zpair
            + self.b2 * (zpair ** 2)
            - self.c1 * zself
            - self.c2 * (zself ** 2)
        )
        return max(a, 0.0)

    def calculate_gvalue(self, data: fwdpy11.PyDiploidGeneticValueData) -> float:
        g = self.phenotypes[data.offspring_metadata.label]
        memoryview(data)[0] = g
        return g

    def sigmoidize(self, g: float) -> float:
        return 1.0 / (1.0 + math.e ** -(self.slope * (g + self.sig0)))

    def phenotype(
        self, pop: fwdpy11.DiploidPopulation, metadata: fwdpy11.DiploidMetadata
    ) -> float:
        """
        An individual's phenotype is determined by additive effects
        plus a sigmoid function that bounds trait values on [0, 1]
        and places a non-mutant individual at trait value self.p0.
        """
        g = fwdpy11.strict_additive_effects(pop, metadata)
        return self.sigmoidize(g)

    def update(self, pop: fwdpy11.DiploidPopulation) -> None:
        self.phenotypes = [self.phenotype(pop, d) for d in pop.diploid_metadata]


class GetP(object):
    def __init__(self):
        self.data = []

    def __call__(self, pop, sampler):
        if pop.generation % 100 == 0.0:
            md = np.array(pop.diploid_metadata, copy=False)
            self.data.append(np.array(md["g"]))


def plot_trait_values(
    vals, bins=100, interpolation="bicubic", vmin=None, vmax=None, tmult=1
):
    parr = np.array(vals)
    nruns = parr.shape[0]

    phist = np.zeros((nruns, bins))
    for i in range(nruns):
        ph, xe = np.histogram(parr[i].ravel(), bins=bins, range=[0.0, 1.0],)
        phist[i] = ph

    plt.imshow(
        phist,
        aspect="auto",
        interpolation=interpolation,
        origin="lower",
        cmap=plt.cm.binary,
        extent=(min(xe), max(xe), 0, nruns),
        vmin=vmin,
        vmax=vmax,
    )
    plt.yticks(
        np.around(np.linspace(0, nruns, 11), -1),
        np.around(np.linspace(0, (nruns - 1) * tmult + 1, 11), -1),
    )
    plt.savefig("pysnowdrift.png")


# Branch point will be at 0.6
gv = PySnowdrift(6.0, -1.4, 4.56, -1.6, 10.0, 0.2, int(sys.argv[2]))


pdict = {
    "nregions": [],
    "sregions": [fwdpy11.GaussianS(0, 1, 1, 5e-3)],
    "recregions": [],
    "rates": (0, 1e-2, None),
    "demography": fwdpy11.DiscreteDemography(),
    "simlen": 40000,
    "gvalue": gv,
    "prune_selected": False,
}


params = fwdpy11.ModelParams(**pdict)
pop = fwdpy11.DiploidPopulation(5000, 1.0)
rng = fwdpy11.GSLrng(int(sys.argv[1]))
r = GetP()
fwdpy11.evolvets(rng, pop, params, 100, r)
plot_trait_values(r.data)



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#


def mslike(pop, **kwargs):
    """
    Function to establish default parameters
    for a single-locus simulation for standard pop-gen
    modeling scenarios.

    :params pop: An instance of :class:`fwdpy11.DiploidPopulation`
    :params kwargs: Keyword arguments.
    """
    import fwdpy11

    if isinstance(pop, fwdpy11.DiploidPopulation) is False:
        raise ValueError("incorrect pop type: " + str(type(pop)))
    defaults = {
        "simlen": 10 * pop.N,
        "beg": 0.0,
        "end": 1.0,
        "theta": 100.0,
        "pneutral": 1.0,
        "rho": 100.0,
        "dfe": None,
    }
    for key, value in kwargs.items():
        if key in defaults:
            defaults[key] = value

    r = defaults["rho"] / (4.0 * float(pop.N))
    params = {
        "simlen": defaults["simlen"],
        "nregions": [fwdpy11.Region(defaults["beg"], defaults["end"], 1.0)],
        "recregions": [fwdpy11.PoissonInterval(defaults["beg"], defaults["end"], r)],
        "rates": (
            (defaults["pneutral"] * defaults["theta"]) / (4.0 * pop.N),
            ((1.0 - defaults["pneutral"]) * defaults["theta"]) / (4.0 * pop.N),
            None
        ),
        "gvalue": fwdpy11.Multiplicative(2.0),
    }
    if defaults["dfe"] is None:
        params["sregions"] = []
    else:
        params["sregions"] = [defaults["dfe"]]
    return params



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import unittest

import fwdpy11
import numpy as np


class TestDiploidPopulation(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.pop = fwdpy11.DiploidPopulation(1000, 1.0)

    def test_N(self):
        self.assertEqual(self.pop.N, 1000)

    def test_founder_genome_counts(self):
        self.assertEqual(1, len(self.pop.haploid_genomes))
        self.assertEqual(2 * self.pop.N, self.pop.haploid_genomes[0].n)

    def test_generation(self):
        self.assertEqual(self.pop.generation, 0)

    def test_fitnesses(self):
        # All fitnesses should be 1
        self.assertEqual(
            sum([i.w for i in self.pop.diploid_metadata]), float(self.pop.N)
        )

    def test_labels(self):
        # All diploids should be labeled 0 to pop.N-1
        self.assertEqual(
            [i.label for i in self.pop.diploid_metadata], [i for i in range(self.pop.N)]
        )

    def test_genetic_values(self):
        self.assertEqual(sum([i.g for i in self.pop.diploid_metadata]), 0.0)

    def test_e_values(self):
        self.assertEqual(sum([i.e for i in self.pop.diploid_metadata]), 0.0)


class TestDiploidPopulationExceptions(unittest.TestCase):
    def testNzero(self):
        with self.assertRaises(ValueError):
            fwdpy11.DiploidPopulation(0, 1.0)


# NOTE: this tests an internal function (pop._record_ancient_samples),
# which is NOT the canonical user-facing API for doing this!
def test_ancient_sample_recording():
    pop = fwdpy11.DiploidPopulation(1000, 1.0)
    pop._record_ancient_samples([i for i in range(pop.N)])
    md = np.array(pop.diploid_metadata, copy=False)
    amd = np.array(pop.ancient_sample_metadata, copy=False)
    assert np.array_equal(md, amd)

    pop = fwdpy11.DiploidPopulation(1000, 1.0)
    pop._record_ancient_samples([i for i in range(0, pop.N, 2)])
    md = np.array(pop.diploid_metadata, copy=False)
    amd = np.array(pop.ancient_sample_metadata, copy=False)
    assert len(amd) == pop.N // 2
    assert np.array_equal(md[::2], amd)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2023 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import demes
import pytest

import fwdpy11


def _validate_expected(regions, numbers):
    assert len([i for i in regions if isinstance(
        i, fwdpy11.PoissonCrossoverGenerator)]) == numbers[0]

    assert len([i for i in regions if isinstance(
        i, fwdpy11.NonPoissonCrossoverGenerator)]) == numbers[1]


def roundtrip(regions):
    yaml = """
    time_units: generations
    demes:
    - name: A
      epochs:
        - {end_time: 0, start_size: 10}
    """
    graph = demes.loads(yaml)
    demog = fwdpy11.discrete_demography.from_demes(graph, burnin=1)
    pdict = {"recregions": regions,
             "nregions": [],
             "sregions": [],
             "rates": (0, 0, None),
             "gvalue": fwdpy11.Multiplicative(2.),
             "simlen": 10,
             "demography": demog,
             }
    params = fwdpy11.ModelParams(**pdict)
    pop = fwdpy11.DiploidPopulation(10, 2)
    rng = fwdpy11.GSLrng(123)
    fwdpy11.evolvets(rng, pop, params, 100)


# The setup is ([regions], (number poisson, number non-poisson))
# The latter is used to mimic the internal idiom that we
# use to set up the genetic map to send to C++
@pytest.mark.parametrize("data", [
    ([fwdpy11.PoissonInterval(0, 1, 1e-3)], (1, 0)),
    ([fwdpy11.PoissonPoint(1, 1e-3)], (1, 0)),
    ([fwdpy11.PoissonInterval(0, 1, 1e-3),
      fwdpy11.PoissonPoint(1, 1e-3)], (2, 0)),
    ([fwdpy11.BinomialInterval(0, 1, 1e-3)], (0, 1)),
    ([fwdpy11.BinomialPoint(1, 1e-3)], (0, 1)),
    ([fwdpy11.BinomialInterval(0, 1, 1e-3),
      fwdpy11.BinomialPoint(1, 1e-3)], (0, 2)),
    ([fwdpy11.PoissonInterval(0, 0.1, 1e-3),
     fwdpy11.PoissonPoint(0.1, 1e-3),
     fwdpy11.BinomialInterval(0.1, 0.2, 1e-3),
     fwdpy11.BinomialPoint(0.2, 1e-3)], (2, 2)),
    ([fwdpy11.BinomialIntervalMap(0.5, [fwdpy11.Region(0, 1, 1e-5),
                                        fwdpy11.Region(1, 2, 2)])], (0, 1))
])
def test_model_params_with_recregions(data):
    regions, expected = data
    _validate_expected(regions, expected)
    roundtrip(regions)



import typing

import attr

from ..class_decorators import (attr_add_asblack, attr_class_pickle,
                                attr_class_to_from_dict_no_recurse)


from .demographic_model_citation import DemographicModelCitation
from .forward_demes_graph import ForwardDemesGraph


@attr_add_asblack
@attr_class_pickle
@attr_class_to_from_dict_no_recurse
@attr.s(
    kw_only=True, frozen=True, auto_attribs=True, repr_ns="fwdpy11"
)
class DemographicModelDetails(object):
    """
    Stores rich information about a demographic model.
    Instances of this class get returned by functions
    generating pre-calculated models.

    This class has the following attributes, whose names
    are also ``kwargs`` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param model: The demographic model parameters
    :type model: fwdpy11.ForwardDemesGraph
    :param name: The name of the model
    :type name: str
    :param source: The source of the model
    :type source: dict
    :param parameters: The parameters used to generate ``model``
    :type parameters: object
    :param citation: Citation information
    :type citation: dict or fwdpy11.DemographicModelCitation or None
    :param metadata: Optional field for additional info
    :type metadata: object

    .. versionadded:: 0.8.0
    """
    model: ForwardDemesGraph
    name: str
    source: typing.Dict
    parameters: object
    citation: typing.Optional[typing.Union[DemographicModelCitation,
                                           typing.Dict]]
    metadata: typing.Optional[object] = None

    @property
    def initial_sizes_list(self) -> typing.List[int]:
        """
        A list of the nonzero (parental) deme sizes at time 0 of the model
        (thinking forwards in time).

        This property is useful for setting up a
        :class:`fwdpy11.DiploidPopulation`.
        """
        return self.model.initial_sizes

    @property
    def total_simulation_length(self) -> int:
        """
        The total number of generations in the model.

        This value includes any burn-in time and starts
        from time 0 (forwards in time).
        """
        return self.metadata["total_simulation_length"]

    @property
    def deme_labels(self) -> typing.Dict[int, str]:
        """
        A dictionary mapping integer deme IDs to deme names
        (strings).
        """
        return self.metadata["deme_labels"]



# A simple parser that builds a fully-qualified Demes Graph from an input JSON
# string.
#
# Requires Python 3.7+.
#
# This implementation is NOT recommended for use in any downstream software and
# is provided purely as reference material for parser writers (i.e., in other
# programming languages). Python users should use the "demes" package in their
# software: https://github.com/popsim-consortium/demes-python
#
# The entry point is the ``parse`` function, which returns a fully-qualified
# Graph. The implementation is written with clarity and correctness as the main
# priorities. Its main purpose is to remove any potential ambiguities that may
# exist in the written specification and to simplify the process of writing
# other parsers. In the interest of simplicity, the parser does not generate
# useful error messages in all cases (but we would hope that practical
# implementations would).
#
# Type annotations are used where they help with readability, but not applied
# exhaustively.
from __future__ import annotations

import math
import numbers
import copy
import pprint
import dataclasses
from typing import Dict, List, Union

# Numerical wiggle room.
EPSILON = 1e-6

# JSON does not provide a way to encode IEEE infinity values, which we
# require to describe start_time values. To work around this we use the
# string "Infinity" to represent IEEE positive infinity.
JSON_INFINITY_STR = "Infinity"


def parse(data: dict) -> Graph:
    # Parsing is done by popping items out of the input data dictionary and
    # creating the appropriate Python objects. We ensure that extra items
    # have not been included in the data payload by checking if the objects
    # are empty once we have removed all the values defined in the
    # specification. Type and range validation of simple items (e.g., the
    # value must be a positive integer) is performed at the same time,
    # using the pop_x functions. Once the full object model of the input
    # data has been built, the rules for creating a fully-qualified Demes
    # graph are applied in the "resolve" functions. Finally, we validate
    # the fully-qualified graph to ensure that relationships between the
    # entities have been specified correctly.
    data = copy.deepcopy(data)

    defaults = pop_object(data, "defaults", {})
    deme_defaults = pop_object(defaults, "deme", {})
    migration_defaults = pop_object(defaults, "migration", {})
    pulse_defaults = pop_object(defaults, "pulse", {})
    # epoch defaults may also be specified within a Deme definition.
    global_epoch_defaults = pop_object(defaults, "epoch", {})
    check_empty(defaults)

    graph = Graph(
        description=pop_string(data, "description", ""),
        time_units=pop_string(data, "time_units", None),
        doi=pop_list(data, "doi", [], str, is_nonempty),
        generation_time=pop_number(
            data, "generation_time", None, is_positive_and_finite
        ),
        metadata=pop_object(data, "metadata", {}),
    )
    check_defaults(
        deme_defaults,
        dict(
            description=(str, None),
            start_time=((str, numbers.Number), is_positive_or_json_infinity),
            ancestors=(list, is_list_of_identifiers),
            proportions=(list, is_list_of_fractions),
        ),
    )

    allowed_epoch_defaults = dict(
        end_time=(numbers.Number, is_non_negative_and_finite),
        start_size=(numbers.Number, is_positive_and_finite),
        end_size=(numbers.Number, is_positive_and_finite),
        selfing_rate=(numbers.Number, is_fraction),
        cloning_rate=(numbers.Number, is_fraction),
        size_function=(str, None),
    )
    check_defaults(global_epoch_defaults, allowed_epoch_defaults)

    for deme_data in pop_list(data, "demes"):
        insert_defaults(deme_data, deme_defaults)
        deme = graph.add_deme(
            name=pop_string(deme_data, "name", validator=is_identifier),
            description=pop_string(deme_data, "description", ""),
            start_time=pop_number(
                deme_data,
                "start_time",
                None,
                is_positive_or_json_infinity,
                allow_inf=True,
            ),
            ancestors=pop_list(deme_data, "ancestors", [], str, is_identifier),
            proportions=pop_list(
                deme_data, "proportions", None, numbers.Number, is_fraction
            ),
        )

        local_defaults = pop_object(deme_data, "defaults", {})
        local_epoch_defaults = pop_object(local_defaults, "epoch", {})
        check_empty(local_defaults)
        check_defaults(local_epoch_defaults, allowed_epoch_defaults)
        epoch_defaults = global_epoch_defaults.copy()
        epoch_defaults.update(local_epoch_defaults)
        check_defaults(epoch_defaults, allowed_epoch_defaults)

        # There is always at least one epoch defined with the default values.
        for epoch_data in pop_list(deme_data, "epochs", [{}]):
            insert_defaults(epoch_data, epoch_defaults)
            deme.add_epoch(
                end_time=pop_number(
                    epoch_data, "end_time", None, is_non_negative_and_finite
                ),
                start_size=pop_number(
                    epoch_data, "start_size", None, is_positive_and_finite
                ),
                end_size=pop_number(
                    epoch_data, "end_size", None, is_positive_and_finite
                ),
                selfing_rate=pop_number(epoch_data, "selfing_rate", 0, is_fraction),
                cloning_rate=pop_number(epoch_data, "cloning_rate", 0, is_fraction),
                size_function=pop_string(epoch_data, "size_function", None),
            )
            check_empty(epoch_data)
        check_empty(deme_data)

        if len(deme.epochs) == 0:
            raise ValueError(f"no epochs for deme {deme.name}")

    if len(graph.demes) == 0:
        raise ValueError("the graph must have one or more demes")

    check_defaults(
        migration_defaults,
        dict(
            rate=(numbers.Number, is_fraction),
            start_time=((numbers.Number, str), is_positive_or_json_infinity),
            end_time=(numbers.Number, is_non_negative_and_finite),
            source=(str, is_identifier),
            dest=(str, is_identifier),
            demes=(list, is_list_of_identifiers),
        ),
    )
    for migration_data in pop_list(data, "migrations", []):
        insert_defaults(migration_data, migration_defaults)
        graph.add_migration(
            rate=pop_number(migration_data, "rate", validator=is_fraction),
            start_time=pop_number(
                migration_data,
                "start_time",
                None,
                is_positive_or_json_infinity,
                allow_inf=True,
            ),
            end_time=pop_number(
                migration_data, "end_time", None, is_non_negative_and_finite
            ),
            source=pop_string(migration_data, "source", None, is_nonempty),
            dest=pop_string(migration_data, "dest", None, is_nonempty),
            demes=pop_list(
                migration_data,
                "demes",
                default=None,
                required_type=str,
                validator=is_identifier,
            ),
        )
        check_empty(migration_data)

    check_defaults(
        pulse_defaults,
        dict(
            sources=(list, is_nonempty_list_of_identifiers),
            dest=(str, is_identifier),
            time=(numbers.Number, is_positive_and_finite),
            proportions=(list, is_nonempty_list_of_fractions_with_sum_less_than_1),
        ),
    )
    for pulse_data in pop_list(data, "pulses", []):
        insert_defaults(pulse_data, pulse_defaults)
        graph.add_pulse(
            sources=pop_list(
                pulse_data,
                "sources",
                default=[],
                required_type=str,
                validator=is_identifier,
            ),
            dest=pop_string(pulse_data, "dest", validator=is_identifier),
            time=pop_number(pulse_data, "time", validator=is_positive_and_finite),
            proportions=pop_list(
                pulse_data,
                "proportions",
                default=[],
                required_type=numbers.Number,
                validator=is_fraction,
            ),
        )
        check_empty(pulse_data)

    check_empty(data)

    # The input object model has now been fully populated, and local type and
    # value checking done. Default values (either from the schema or set explicitly
    # by the user via "defaults" sections) have been assigned. We now "resolve"
    # the model so that any values that can be imputed from the structure of the
    # model are set explicitly. Once this is done, we then validate the model to
    # check that the relationships between various entities make sense. Note that
    # there isn't a clean separation between resolution and validation here, since
    # some validation is simplest to perform as part of the resolution logic in
    # this particular implementation.
    graph.resolve()
    graph.validate()

    return graph


def encode_inf(value):
    if math.isinf(value):
        return JSON_INFINITY_STR
    return value


# Validator functions. These are used as arguments to the pop_x functions and
# check properties of the values.


def is_positive_or_json_infinity(value):
    return value == JSON_INFINITY_STR or value > 0


def is_positive_and_finite(value):
    return value > 0 and not math.isinf(value)


def is_non_negative_and_finite(value):
    return value >= 0 and not math.isinf(value)


def is_fraction(value):
    return 0 <= value <= 1


def is_nonempty(value):
    return len(value) > 0


def is_identifier(value):
    return value.isidentifier()


def is_list_of_identifiers(value):
    return all(isinstance(v, str) and is_identifier(v) for v in value)


def is_nonempty_list_of_identifiers(value):
    return is_list_of_identifiers(value) and len(value) > 0


def is_list_of_fractions(value):
    return all(isinstance(v, numbers.Number) and is_fraction(v) for v in value)


def is_nonempty_list_of_fractions_with_sum_less_than_1(value):
    return is_list_of_fractions(value) and len(value) > 0 and sum(value) <= 1


def validate_item(name, value, required_type, validator=None):
    if not isinstance(value, required_type):
        raise TypeError(
            f"Attribute '{name}' must be a {required_type}; "
            f"current type is {type(value)}."
        )
    if validator is not None and not validator(value):
        validator_name = validator.__name__[3:]  # Strip off is_ from function name
        raise ValueError(f"Attribute '{name}' is not {validator_name}")


# We need to use this trick because None is a meaningful input value for these
# pop_x functions.
NO_DEFAULT = object()


def pop_item(data, name, *, required_type, default=NO_DEFAULT, validator=None):
    if name in data:
        value = data.pop(name)
        validate_item(name, value, required_type, validator)
    else:
        if default is NO_DEFAULT:
            raise KeyError(f"Attribute '{name}' is required")
        value = default
    return value


def pop_list(data, name, default=NO_DEFAULT, required_type=None, validator=None):
    value = pop_item(data, name, default=default, required_type=list)
    if required_type is not None and value is not None:
        for item in value:
            validate_item(name, item, required_type, validator)
    return value


def pop_object(data, name, default=NO_DEFAULT):
    return pop_item(data, name, default=default, required_type=dict)


def pop_string(data, name, default=NO_DEFAULT, validator=None):
    return pop_item(data, name, default=default, required_type=str, validator=validator)


def pop_number(data, name, default=NO_DEFAULT, validator=None, allow_inf=False):
    # If infinite values are allowed for this number, the string "Infinity"
    # is also accepted, and so str is an accepted type. There is a small loophole
    # here in which string numbers like "1000" will be accepted by the type
    # checking machinery, but the is_positive_or_json_infinity validator
    # will catch this and raise a TypeError when it tries to compare with 0.
    if allow_inf:
        assert validator is is_positive_or_json_infinity
    required_type = (numbers.Number, str) if allow_inf else numbers.Number
    value = pop_item(
        data,
        name,
        default=default,
        required_type=required_type,
        validator=validator,
    )
    if value == JSON_INFINITY_STR:
        return math.inf
    return value


def check_empty(data):
    if len(data) != 0:
        raise ValueError(f"Extra fields are not permitted:{data}")


def check_defaults(defaults, allowed_fields):
    for key, value in defaults.items():
        if key not in allowed_fields:
            raise ValueError(
                f"Only fields {list(allowed_fields.keys())} can be specified "
                "in the defaults"
            )
        required_type, validator = allowed_fields[key]
        validate_item(key, value, required_type, validator)


def insert_defaults(data, defaults):
    for key, value in defaults.items():
        if key not in data:
            data[key] = value


@dataclasses.dataclass
class Interval:
    """
    A half-open time interval (start_time, end_time].
    """

    start_time: float
    end_time: float

    def __init__(self, start_time, end_time):
        assert start_time > end_time
        self.start_time = start_time
        self.end_time = end_time

    def intersects(self, other):
        """True if self and other intersect, False otherwise."""
        assert isinstance(other, self.__class__)
        return not (
            self.end_time >= other.start_time or other.end_time >= self.start_time
        )

    def is_subinterval(self, other):
        """True if self is completely contained within other, False otherwise."""
        assert isinstance(other, self.__class__)
        return self.start_time <= other.start_time and self.end_time >= other.end_time

    def __contains__(self, time):
        return self.start_time > time >= self.end_time


@dataclasses.dataclass
class Epoch:
    end_time: Union[float, None]
    start_size: Union[float, None]
    end_size: Union[float, None]
    size_function: str
    selfing_rate: float
    cloning_rate: float

    def as_json_dict(self) -> dict:
        return dataclasses.asdict(self)

    def resolve(self):
        if self.size_function is None:
            if self.start_size == self.end_size:
                self.size_function = "constant"
            else:
                self.size_function = "exponential"

    def validate(self):
        if self.size_function not in ("constant", "exponential", "linear"):
            raise ValueError(f"unknown size_function '{self.size_function}'")
        if self.size_function == "constant" and self.start_size != self.end_size:
            raise ValueError(
                "size_function is constant but "
                f"start_size ({self.start_size}) != end_size ({self.end_size})"
            )


@dataclasses.dataclass
class Deme:
    name: str
    start_time: Union[None, float]
    description: str
    ancestors: List[Deme]
    proportions: Union[List[float], None]
    epochs: List[Epoch] = dataclasses.field(default_factory=list)

    def add_epoch(
        self,
        end_time: Union[float, None],
        start_size: Union[float, None],
        end_size: Union[float, None],
        selfing_rate: float,
        cloning_rate: float,
        size_function: str,
    ) -> Epoch:
        epoch = Epoch(
            end_time=end_time,
            start_size=start_size,
            end_size=end_size,
            selfing_rate=selfing_rate,
            cloning_rate=cloning_rate,
            size_function=size_function,
        )
        self.epochs.append(epoch)
        return epoch

    @property
    def end_time(self):
        return self.epochs[-1].end_time

    @property
    def time_interval(self):
        return Interval(self.start_time, self.end_time)

    def as_json_dict(self) -> dict:
        return {
            "name": self.name,
            "description": self.description,
            "start_time": encode_inf(self.start_time),
            "epochs": [epoch.as_json_dict() for epoch in self.epochs],
            "proportions": self.proportions,
            "ancestors": [deme.name for deme in self.ancestors],
        }

    def __resolve_times(self):
        if self.start_time is None:
            default = math.inf
            if len(self.ancestors) == 1:
                default = self.ancestors[0].epochs[-1].end_time
            elif len(self.ancestors) > 1:
                raise ValueError(
                    "Must explicitly set Deme.start_time when > 1 ancestor"
                )
            self.start_time = default
        if len(self.ancestors) == 0 and not math.isinf(self.start_time):
            raise ValueError(
                f"deme {self.name} has finite start_time, but no ancestors"
            )

        for ancestor in self.ancestors:
            if self.start_time not in ancestor.time_interval:
                raise ValueError(
                    f"Deme {ancestor.name} ({ancestor.time_interval}) doesn't "
                    f"exist at deme {self.name}'s start_time ({self.start_time})"
                )

        # The last epoch has a default end_time of 0
        last_epoch = self.epochs[-1]
        if last_epoch.end_time is None:
            last_epoch.end_time = 0
        last_time = self.start_time
        for epoch in self.epochs:
            if epoch.end_time is None:
                raise ValueError("Epoch end_time must be specified")
            if epoch.end_time >= last_time:
                raise ValueError("Epoch end_times must be in decreasing order.")
            last_time = epoch.end_time

    def __resolve_sizes(self):
        first_epoch = self.epochs[0]
        # The first epoch must specify either start_size or end_size
        if first_epoch.start_size is None and first_epoch.end_size is None:
            raise ValueError(
                "Must specify one or more of start_size and end_size "
                "for the initial epoch"
            )
        if first_epoch.start_size is None:
            first_epoch.start_size = first_epoch.end_size
        if first_epoch.end_size is None:
            first_epoch.end_size = first_epoch.start_size
        last_epoch = first_epoch
        for epoch in self.epochs[1:]:
            if epoch.start_size is None:
                epoch.start_size = last_epoch.end_size
            if epoch.end_size is None:
                epoch.end_size = epoch.start_size
            last_epoch = epoch

        if self.start_time == math.inf:
            if first_epoch.start_size != first_epoch.end_size:
                raise ValueError(
                    "Cannot have varying population size in an infinite time interval"
                )

    def __resolve_proportions(self):
        if self.proportions is None:
            if len(self.ancestors) == 0:
                self.proportions = []
            elif len(self.ancestors) == 1:
                self.proportions = [1]
            else:
                raise ValueError("Must specify proportions for > 1 ancestor demes")

    def resolve(self):
        self.__resolve_times()
        self.__resolve_sizes()
        self.__resolve_proportions()
        for epoch in self.epochs:
            epoch.resolve()

    def validate(self):
        if len(self.proportions) != len(self.ancestors):
            raise ValueError("proportions must be same length as ancestors")
        if len(self.ancestors) > 0:
            if not math.isclose(sum(self.proportions), 1):
                raise ValueError("Sum of proportions must be approximately 1")
        if len(set(anc.name for anc in self.ancestors)) != len(self.ancestors):
            raise ValueError("ancestors list contains duplicates")
        for epoch in self.epochs:
            epoch.validate()


@dataclasses.dataclass
class Pulse:
    sources: List[Deme]
    dest: Deme
    time: float
    proportions: List[float]

    def as_json_dict(self) -> dict:
        d = dataclasses.asdict(self)
        d["sources"] = [source.name for source in self.sources]
        d["dest"] = self.dest.name
        return d

    def validate(self):
        sources_names = set(source.name for source in self.sources)
        if self.dest.name in sources_names:
            raise ValueError("Cannot have source deme equal to dest")
        if len(sources_names) != len(self.sources):
            raise ValueError("Duplicate deme in sources")
        if len(self.sources) == 0:
            raise ValueError("Must have one or more source demes")
        if len(self.sources) != len(self.proportions):
            raise ValueError("Sources and proportions must have same lengths")
        for source in self.sources:
            if self.time not in source.time_interval:
                raise ValueError(
                    f"Deme {source.name} does not exist at time {self.time}"
                )
        # Time limits for the destination deme are different to the source deme,
        # because the destination deme is affected immediately after the time
        # of the pulse. Thus, a pulse can occur at the destination deme's
        # start_time, but not at the destination deme's end_time.
        if not (self.dest.start_time >= self.time > self.dest.end_time):
            raise ValueError(
                f"Deme {self.dest.name} does not exist at time {self.time}"
            )

        if sum(self.proportions) > 1 + EPSILON:
            raise ValueError(
                f"Pulse proportions into {self.dest.name} at time {self.time} "
                "sum to more than 1"
            )


@dataclasses.dataclass
class Migration:
    rate: Union[float, None]
    start_time: Union[float, None]
    end_time: Union[float, None]
    source: Deme
    dest: Deme

    @property
    def time_interval(self):
        return Interval(self.start_time, self.end_time)

    def as_json_dict(self) -> dict:
        d = dataclasses.asdict(self)
        d["start_time"] = encode_inf(self.start_time)
        d["source"] = self.source.name
        d["dest"] = self.dest.name
        return d

    def resolve(self):
        if self.start_time is None:
            self.start_time = min(self.source.start_time, self.dest.start_time)
        if self.end_time is None:
            self.end_time = max(self.source.end_time, self.dest.end_time)

    def validate(self):
        if self.start_time <= self.end_time:
            raise ValueError("start_time must be > end_time")
        if self.source.name == self.dest.name:
            raise ValueError("Cannot migrate from a deme to itself")
        for deme in [self.source, self.dest]:
            if not self.time_interval.is_subinterval(deme.time_interval):
                raise ValueError(
                    "Migration time interval must be within the each deme's "
                    "time interval"
                )


@dataclasses.dataclass
class Graph:
    time_units: str
    generation_time: Union[float, None]
    doi: List[str]
    description: str
    metadata: dict
    demes: Dict[str, Deme] = dataclasses.field(default_factory=dict)
    migrations: List[Migration] = dataclasses.field(default_factory=list)
    pulses: List[Pulse] = dataclasses.field(default_factory=list)

    def add_deme(
        self,
        name: str,
        description: str,
        start_time: Union[float, None],
        ancestors: List[str],
        proportions: Union[List[float], None],
    ) -> Deme:
        deme = Deme(
            name=name,
            description=description,
            start_time=start_time,
            ancestors=[self.demes[deme_name] for deme_name in ancestors],
            proportions=proportions,
        )
        if deme.name in self.demes:
            raise ValueError(f"Duplicate deme name '{deme.name}'")
        self.demes[deme.name] = deme
        return deme

    def add_migration(
        self,
        *,
        rate: float,
        start_time: Union[float, None],
        end_time: Union[float, None],
        source: Union[str, None],
        dest: Union[str, None],
        demes: Union[List[str], None],
    ) -> List[Migration]:
        migrations: List[Migration] = []
        if not (
            # symmetric
            (demes is not None and source is None and dest is None)
            # asymmetric
            or (demes is None and source is not None and dest is not None)
        ):
            raise ValueError("Must specify either source and dest, or demes")
        if source is not None:
            assert dest is not None
            migrations.append(
                Migration(
                    rate=rate,
                    start_time=start_time,
                    end_time=end_time,
                    source=self.demes[source],
                    dest=self.demes[dest],
                )
            )
        else:
            assert demes is not None
            if len(demes) < 2:
                raise ValueError("Must specify two or more deme names")
            for j, deme_a in enumerate(demes, 1):
                for deme_b in demes[j:]:
                    migration_ab = Migration(
                        rate=rate,
                        start_time=start_time,
                        end_time=end_time,
                        source=self.demes[deme_a],
                        dest=self.demes[deme_b],
                    )
                    migration_ba = Migration(
                        rate=rate,
                        start_time=start_time,
                        end_time=end_time,
                        source=self.demes[deme_b],
                        dest=self.demes[deme_a],
                    )
                    migrations.extend([migration_ab, migration_ba])
        self.migrations.extend(migrations)
        return migrations

    def add_pulse(
        self, sources: List[str], dest: str, time: float, proportions: List[float]
    ):
        pulse = Pulse(
            sources=[self.demes[source] for source in sources],
            dest=self.demes[dest],
            time=time,
            proportions=proportions,
        )
        self.pulses.append(pulse)
        return pulse

    def __str__(self):
        data = self.as_json_dict()
        return pprint.pformat(data, indent=2)

    def as_json_dict(self):
        d = dataclasses.asdict(self)
        d["demes"] = [deme.as_json_dict() for deme in self.demes.values()]
        d["migrations"] = [migration.as_json_dict() for migration in self.migrations]
        d["pulses"] = [pulse.as_json_dict() for pulse in self.pulses]
        return d

    def validate(self):
        if self.generation_time is None:
            if self.time_units == "generations":
                self.generation_time = 1
            else:
                raise ValueError(
                    "Must specify Graph.generation_time if time_units is not "
                    "'generations'"
                )
        if self.time_units == "generations" and self.generation_time != 1:
            raise ValueError(
                "If time_units are in generations, generation_time must be 1"
            )
        for deme in self.demes.values():
            deme.validate()
        for pulse in self.pulses:
            pulse.validate()
        for migration in self.migrations:
            migration.validate()

        # Migrations involving the same source and dest can't overlap temporally.
        for j, migration_a in enumerate(self.migrations, 1):
            for migration_b in self.migrations[j:]:
                if (
                    migration_a.source == migration_b.source
                    and migration_a.dest == migration_b.dest
                    and migration_a.time_interval.intersects(migration_b.time_interval)
                ):
                    start_time = min(migration_a.end_time, migration_b.end_time)
                    end_time = max(migration_a.start_time, migration_b.start_time)
                    raise ValueError(
                        f"Competing migration definitions for {migration_a.source.name} "
                        f"and {migration_a.dest.name} during time interval "
                        f"({start_time}, {end_time}]"
                    )

        # The rate of migration entering a deme cannot be more than 1 in any
        # given interval of time.
        time_boundaries = set()
        time_boundaries.update(migration.start_time for migration in self.migrations)
        time_boundaries.update(migration.end_time for migration in self.migrations)
        time_boundaries.discard(math.inf)
        end_times = sorted(time_boundaries, reverse=True)
        start_times = [math.inf] + end_times[:-1]
        ingress_rates = {deme_name: [0.0] * len(end_times) for deme_name in self.demes}
        for j, (start_time, end_time) in enumerate(zip(start_times, end_times)):
            current_interval = Interval(start_time, end_time)
            for migration in self.migrations:
                if current_interval.intersects(migration.time_interval):
                    rate = ingress_rates[migration.dest.name][j] + migration.rate
                    if rate > 1 + EPSILON:
                        raise ValueError(
                            f"Migration rates into {migration.dest.name} sum to "
                            "more than 1 during the time inverval "
                            f"({start_time}, {end_time}]"
                        )
                    ingress_rates[migration.dest.name][j] = rate

    def resolve(self):
        # A deme's ancestors must be listed before it, so any deme we
        # visit must always be visited after its ancestors.
        for deme in self.demes.values():
            deme.resolve()
        for migration in self.migrations:
            migration.resolve()

        # Sort pulses from oldest to youngest.
        # In a discrete-time setting, non-integer pulse times that are distinct
        # could be rounded to the same time value. If the input file has the pulses
        # in time-ascending order, then the pulses would occur in the opposite order
        # compared to a continuous-time simulator. Sorting before the rounding
        # occurs avoids this ambiguity, so we explicitly require pulses to be
        # sorted as part of the parser.
        # Note that Python implements "stable" sorting, which maintains the order
        # of pulses that have the same time value to start with (as required by
        # the spec).
        self.pulses.sort(key=lambda pulse: pulse.time, reverse=True)



#
# Copyright (C) 2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Defining models with specific outcomes.

The API of this module may be subjet to change.
Hence, the documentation is minimal.
"""

import typing
from enum import Enum

import attr
import fwdpy11
from fwdpy11.class_decorators import attr_class_to_from_dict


class AddMutationFailure(Exception):
    pass


class OutOfAttempts(Exception):
    pass


class AncientSamplePolicy(Enum):
    """
    When to record "preserved" or "ancient" samples:
    """

    NEVER = 0
    """
    Never.
    """
    DURATION = 1
    """
    During the entire sojourn of the mutation from when it is first
    added to when the terminiation condition is first met.
    """
    COMPLETION = 2
    """
    Only record the generation when the terminiation condition is first met.
    """


def _non_negative_value(_, attribute, value):
    if value < 0:
        raise ValueError(f"{attribute.name} must be >= 0, got {value}")


def _greater_than_zero(_, attribute, value):
    if value < 0:
        raise ValueError(f"{attribute.name} must be > 0")


def _maximum_greater_minimum(instance, attribute, value):
    if value <= instance.minimum:
        raise ValueError(f"{attribute.name} must be > minimum value")


def _right_greater_left(instance, attribute, value):
    if value <= instance.left:
        raise ValueError(f"{attribute.name} must be > left value")


def _is_polymorphic_frequency(_, attribute, value):
    if not 0.0 < value < 1.0:
        raise ValueError(
            f"{attribute.name} must be 0.0 < {attribute.name} < 1.0")


@attr.s(auto_attribs=True, frozen=True)
class AlleleCount:
    """
    Specify a *number* of copies of a mutation.

    :param count: Initial number of copies of a mutation.
                  This value must be `> 0`.
    :type count: int
    """

    count: int = attr.ib(
        validator=[attr.validators.instance_of(int), _greater_than_zero]
    )


@attr.s(auto_attribs=True, frozen=True)
class AlleleCountRange:
    """
    Specify a range for a *number* of copies of a mutation.

    :param minimum: Minimum number of copies of a mutation.
                    This value must be `> 0`.
    :type minimum: int
    :param maximum: Maximum number of copies of a mutation.
                    This value must be `> minimum`.
    :type maximum: int
    """

    minimum: int = attr.ib(
        validator=[attr.validators.instance_of(int), _greater_than_zero]
    )
    maximum: int = attr.ib(
        validator=[
            attr.validators.instance_of(int),
            _greater_than_zero,
            _maximum_greater_minimum,
        ]
    )


@attr.s(auto_attribs=True, frozen=True)
class FrequencyRange:
    """
    Specify a range for the initial *frequency* of a mutation

    :param minimum: Minimum frequency of a mutation.
                    This value must be `> 0.0` and `< 1.0`.
    :type minimum: float
    :param maximum: Maximum frequency of a mutation.
                    This value must be `> minimum`.
    :type maximum: float
    """

    minimum: float = attr.ib(
        validator=[attr.validators.instance_of(
            float), _is_polymorphic_frequency]
    )
    maximum: float = attr.ib(
        validator=[
            attr.validators.instance_of(float),
            _is_polymorphic_frequency,
            _maximum_greater_minimum,
        ]
    )


@attr.s(auto_attribs=True, kw_only=True, frozen=True)
class PositionRange:
    """
    Specify a half-open interval within which to place a mutation.

    :param left: The left edge, inclusive, of the interval
    :type left: float
    :param right: The right edge, exclusive, of the interval
    :type right: float
    """

    left: float = attr.ib(
        validator=[attr.validators.instance_of(float), _non_negative_value]
    )
    right: float = attr.ib(
        validator=[
            attr.validators.instance_of(float),
            _greater_than_zero,
            _right_greater_left,
        ]
    )


@attr.s(auto_attribs=True, kw_only=True, frozen=True)
class NewMutationParameters:
    """
    Details of a new mutation to add to a population.

    Class instances are created via keyword arguments that
    become attribute names:

    :param deme: The id of the deme in which to add the mutation
    :type deme: Optional[int]
    :param frequency: The frequency of the new mutation.
    :type frequency: Union[AlleleCount, AlleleCountRange, FrequencyRange]
    :param position: Where to put the new mutation
    :type position: PositionRange
    :param data: The specifics of the new mutation
    :type data: :class:`fwdpy11.NewMutationData`
    """

    deme: typing.Optional[int] = None
    frequency: typing.Union[AlleleCount, AlleleCountRange, FrequencyRange] = attr.ib(
        validator=attr.validators.instance_of(
            (AlleleCount, AlleleCountRange, FrequencyRange)
        )
    )
    position: PositionRange = attr.ib(
        validator=attr.validators.instance_of(PositionRange)
    )
    data: fwdpy11.NewMutationData = attr.ib(
        validator=attr.validators.instance_of(fwdpy11.NewMutationData)
    )


class SimulationStatus(Enum):
    Restart = 0
    Continue = 1
    Success = 2


@attr.s(auto_attribs=True, kw_only=True)
class ConditionalModelOutput:
    pop: fwdpy11.DiploidPopulation
    """The population with the added mutation"""
    params: fwdpy11.ModelParams
    """The evolved model parameters"""
    mutation_index: typing.Optional[int]
    """The index of the new mutation in pop.mutations"""
    fixation_index: typing.Optional[int]
    """The index of the new mutation in pop.fixations"""
    num_descendant_nodes: int
    """The number of alive nodes initially containing the new mutation"""


class GlobalFixation(object):
    """
    A terminiation condition monitoring for global fixation of
    the mutation.
    """

    def __call__(
        self, pop, index: int, key: tuple
    ) -> SimulationStatus:
        if pop.mcounts[index] == 0:
            return SimulationStatus.Restart
        if pop.mcounts[index] == 2 * pop.N:
            return SimulationStatus.Success
        return SimulationStatus.Continue


@attr.s(auto_attribs=True, frozen=True)
class FocalDemeFixation:
    """
    A terminiation condition checking for fixation in a specific deme.
    """

    deme: int = attr.ib(
        validator=[attr.validators.instance_of(int), _non_negative_value]
    )

    def __call__(self, pop, index, key) -> SimulationStatus:
        deme_sizes = pop.deme_sizes(as_dict=True)
        if self.deme not in deme_sizes:
            return SimulationStatus.Restart
        count = 0

        # TODO: we can probably do better here
        # by using the numpy interface to the metadata
        for md in pop.diploid_metadata:
            if md.deme == self.deme:
                for genome in [
                    pop.diploids[md.label].first,
                    pop.diploids[md.label].second,
                ]:
                    if index in pop.haploid_genomes[genome].smutations:
                        count += 1

        if count == 0:
            return SimulationStatus.Restart
        if count == 2 * deme_sizes[self.deme]:
            return SimulationStatus.Success
        return SimulationStatus.Continue


@attr_class_to_from_dict
@attr.s(frozen=True)
class EvolveOptions:
    """
    Options to pass on too :func:`fwdpy11.evolvets`.
    """

    simplification_interval: int = attr.ib(
        validator=[attr.validators.instance_of(int), _non_negative_value], default=100
    )
    suppress_table_indexing: bool = attr.ib(
        validator=attr.validators.instance_of(bool), default=True
    )
    record_gvalue_matrix: bool = attr.ib(
        validator=attr.validators.instance_of(bool), default=False
    )
    preserve_first_generation: bool = attr.ib(
        validator=attr.validators.instance_of(bool), default=False
    )


from ._selective_sweep import _selective_sweep as selective_sweep  # NOQA
from ._track_added_mutation import _track_added_mutation as track_added_mutation  # NOQA



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import enum
import typing
import warnings

import attr
import numpy as np

from ._fwdpy11 import (GeneticValueIsTrait, GeneticValueNoise, _ll_Additive,
                       _ll_GaussianNoise, _ll_GBR,
                       _ll_GaussianStabilizingSelection,
                       _ll_GSSmo,
                       _ll_Multiplicative, _ll_MultivariateGSSmo, _ll_NoNoise,
                       _ll_Optimum, _ll_PleiotropicOptima,
                       _ll_StrictAdditiveMultivariateEffects)
from .class_decorators import (attr_add_asblack, attr_class_pickle_with_super,
                               attr_class_to_from_dict,
                               attr_class_to_from_dict_no_recurse)


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class Optimum(_ll_Optimum):
    """
    Parameters for a trait optimum.

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param optimum: The trait value
    :type optimum: float
    :param VS: Strength of stabilizing selection
    :type VS: float
    :param when: The time when the optimum shifts
    :type when: int or None

    .. versionadded:: 0.7.1

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    optimum: float = attr.ib(validator=attr.validators.instance_of(float))
    VS: float = attr.ib(validator=attr.validators.instance_of(float))
    when: typing.Optional[int] = attr.ib(default=None)

    @when.validator
    def validate_when(self, attribute, value):
        if value is not None:
            attr.validators.instance_of(int)(self, attribute, value)

    def __attrs_post_init__(self):
        super(Optimum, self).__init__(self.optimum, self.VS, self.when)


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11", eq=False)
class PleiotropicOptima(_ll_PleiotropicOptima):
    """
    Parameters for multiple trait optima

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param optima: The trait values
    :type optima: List[float]
    :param VS: Strength of stabilizing selection
    :type VS: float
    :param when: The time when the optimum shifts
    :type when: int or None

    .. versionadded:: 0.7.1

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    optima: typing.List[float]
    VS: float = attr.ib(validator=attr.validators.instance_of(float))
    when: typing.Optional[int] = attr.ib(default=None)

    @when.validator
    def validate_when(self, attribute, value):
        if value is not None:
            attr.validators.instance_of(int)(self, attribute, value)

    def __attrs_post_init__(self):
        super(PleiotropicOptima, self).__init__(self.optima,
                                                self.VS,
                                                self.when)

    def __eq__(self, other):
        optima_equal = np.array_equal(self.optima, other.optima)
        VS_equal = self.VS == other.VS
        when_equal = False
        if self.when is not None and other.when is not None:
            when_equal = self.when == other.when

        return optima_equal and VS_equal and when_equal


@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class GaussianStabilizingSelection(_ll_GaussianStabilizingSelection):
    """
    Define a mapping of phenotype-to-fitness according to a 
    Gaussian stabilizing selection model.

    Instances of this trait must be constructed by one of the
    various class methods available.
    """
    is_single_trait: bool
    optima: list

    def __attrs_post_init__(self):
        if self.optima != sorted(self.optima, key=lambda x: x.when):
            raise ValueError("optima must be sorted by time from past to present")
        if self.is_single_trait is True:
            inner = _ll_GSSmo(self.optima)
        else:
            inner = _ll_MultivariateGSSmo(self.optima)
        super(GaussianStabilizingSelection, self).__init__(inner)

    def __getstate__(self):
        return self.asdict()

    def __setstate__(self, d):
        self.__dict__.update(**d)
        self.__attrs_post_init__()

    @classmethod
    def single_trait(cls, optima: typing.List[Optimum]):
        """
        Stabilizing selection on a single trait

        :param optima: The optimum values.
                       Multiple values specify a moving optimum.
        :type optima: List[fwdpy11.Optimum]

        """
        new_optima = []
        for o in optima:
            if o.when is None:
                new_optima.append(Optimum(o.optimum, o.VS, 0))
            else:
                new_optima.append(o)
        return cls(is_single_trait=True, optima=new_optima)

    @classmethod
    def pleiotropy(cls, optima: typing.List[PleiotropicOptima]):
        """
        Stabilizing selection with pleiotropy.

        :param optima: The optimum values.
                       Multiple values specify a moving optimum.
        :type optima: List[fwdpy11.PleiotropicOptima]
        """
        return cls(is_single_trait=False, optima=optima)

    def asdict(self):
        return {"is_single_trait": self.is_single_trait, "optima": self.optima}

    @classmethod
    def fromdict(cls, data):
        return cls(**data)


@attr_add_asblack
@attr_class_to_from_dict_no_recurse
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class GSS(_ll_GSSmo):
    """
    Gaussian stabilizing selection on a single trait.

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param optimum: The optimal trait value
    :type optimum: float or fwdpy11.Optimum
    :param VS: Inverse strength of stabilizing selection
    :type VS: float or None

    .. note::

        VS should be None if optimum is an instance
        of :class:`fwdpy11.Optimum`

    .. versionchanged:: 0.7.1

        Allow instances of fwdpy11.Optimum for intitialization

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    optimum: typing.Union[Optimum, float]
    VS: typing.Optional[float] = None

    def __attrs_post_init__(self):
        warnings.warn("use GaussianStabilizingSelection instead",
                      DeprecationWarning)
        if self.VS is None:
            super(GSS, self).__init__(
                [Optimum(optimum=self.optimum.optimum,
                         VS=self.optimum.VS, when=0)]
            )
        else:
            super(GSS, self).__init__(
                [Optimum(optimum=self.optimum, VS=self.VS, when=0)]
            )

    def __getstate__(self):
        return self.asdict()

    def __setstate__(self, d):
        self.__dict__.update(d)
        if self.VS is None:
            super(GSS, self).__init__(
                [Optimum(optimum=self.optimum.optimum,
                         VS=self.optimum.VS, when=0)]
            )
        else:
            super(GSS, self).__init__(
                [Optimum(optimum=self.optimum, VS=self.VS, when=0)]
            )


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict_no_recurse
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class GSSmo(_ll_GSSmo):
    """
    Gaussian stabilizing selection on a single trait with moving
    optimum.

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param optima: The optimal trait values
    :type optima: list[fwdpy11.Optimum]

    .. note::

        Instances of fwdpy11.Optimum must have valid
        values for `when`.

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class

    """

    optima: typing.List[Optimum] = attr.ib()

    @optima.validator
    def validate_optima(self, attribute, value):
        if len(value) == 0:
            raise ValueError("list of optima cannot be empty")
        for o in value:
            if o.when is None:
                raise ValueError("Optimum.when is None")

    def __attrs_post_init__(self):
        warnings.warn("use GaussianStabilizingSelection instead",
                      DeprecationWarning)
        super(GSSmo, self).__init__(self.optima)


@attr_add_asblack
@attr_class_to_from_dict_no_recurse
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class MultivariateGSS(_ll_MultivariateGSSmo):
    """
    Multivariate gaussian stablizing selection.

    Maps a multidimensional trait to fitness using the Euclidian
    distance of a vector of trait values to a vector of optima.

    Essentially, this is Equation 1 of

    Simons, Yuval B., Kevin Bullaughey, Richard R. Hudson, and Guy Sella. 2018.
    "A Population Genetic Interpretation of GWAS Findings
    for Human Quantitative Traits."
    PLoS Biology 16 (3): e2002985.

    For the case of moving optima, see :class:`fwdpy11.MultivariateGSSmo`.
    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param optima: The optimum value for each trait over time
    :type optima: numpy.ndarray or list[fwdpy11.PleiotropicOptima]
    :param VS: Inverse strength of stablizing selection
    :type VS: float or None

    .. note::

        `VS` should be `None` if `optima` is list[fwdpy11.PleiotropicOptima]

        `VS` is :math:`\\omega^2` in the Simons et al. notation

    .. versionchanged:: 0.7.1
        Allow initialization with list of fwdpy11.PleiotropicOptima

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    optima: typing.Union[PleiotropicOptima, typing.List[float]]
    VS: typing.Optional[float] = None

    def __attrs_post_init__(self):
        warnings.warn("use GaussianStabilizingSelection instead",
                      DeprecationWarning)
        if self.VS is None:
            super(MultivariateGSS, self).__init__([self.optima])
        else:
            super(MultivariateGSS, self).__init__(self._convert_to_list())

    def __getstate__(self):
        return self.asdict()

    def __setstate__(self, d):
        self.__dict__.update(d)
        if self.VS is None:
            super(MultivariateGSS, self).__init__([self.optima])
        else:
            super(MultivariateGSS, self).__init__(self._convert_to_list())

    def _convert_to_list(self):
        if self.VS is None:
            raise ValueError("VS must not be None")
        return [PleiotropicOptima(optima=self.optima, VS=self.VS, when=0)]


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict_no_recurse
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class MultivariateGSSmo(_ll_MultivariateGSSmo):
    """
    Multivariate gaussian stablizing selection with moving optima
    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param optima: list of optima over time
    :type optima: list[fwdpy11.PleiotropicOptima]

    .. versionchanged:: 0.7.1

        Allow initialization with list of fwdpy11.PleiotropicOptima

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    optima: typing.List[PleiotropicOptima] = attr.ib()

    @optima.validator
    def validate_optima(self, attribute, value):
        if len(value) == 0:
            raise ValueError("list of optima cannot be empty")
        for o in value:
            if o.when is None:
                raise ValueError("PleiotropicOptima.when is None")

    def __attrs_post_init__(self):
        warnings.warn("use GaussianStabilizingSelection instead",
                      DeprecationWarning)
        super(MultivariateGSSmo, self).__init__(self.optima)


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class NoNoise(_ll_NoNoise):
    """
    No random effects on genetic values

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    def __attrs_post_init__(self):
        super(NoNoise, self).__init__()


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class GaussianNoise(_ll_GaussianNoise):
    """
    Gaussian noise added to genetic values.
    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param sd: Standard deviation
    :type sd: float
    :param mean: Mean value
    :type mean: float

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    sd: float
    mean: float = 0.0

    def __attrs_post_init__(self):
        super(GaussianNoise, self).__init__(self.sd, self.mean)


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict_no_recurse
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class Additive(_ll_Additive):
    """
    Additive effects on genetic values.

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param scaling: How to treat mutant homozygotes.
    :type scaling: float
    :param gvalue_to_fitness: How to map trait value to fitness
    :type gvalue_to_fitness: fwdpy11.GeneticValueIsTrait
    :param noise: Random effects on trait values
    :type noise: fwdpy11.GeneticValueNoise

    When `gvalue_to_fitness` is `None`, then we are
    modeling additive effects on fitness.

    For a model of fitness, the genetic value is 1, 1+e*h,
    1+`scaling`*e for genotypes AA, Aa, and aa, respectively,
    where `e` and `h` are the effect size and dominance, respectively.

    For a model of a trait (phenotype), meaning `gvalue_to_fitness`
    is not `None`, the values for the three genotypes are 0, e*h,
    and e, respectively.

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    scaling: float
    gvalue_to_fitness: GeneticValueIsTrait = None
    noise: GeneticValueNoise = None
    ndemes: int = 1

    def __attrs_post_init__(self):
        super(Additive, self).__init__(
            self.scaling, self.gvalue_to_fitness, self.noise, self.ndemes
        )


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict_no_recurse
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class Multiplicative(_ll_Multiplicative):
    """
    Multiplicative effects on genetic values.

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param scaling: How to treat mutant homozygotes.
    :type scaling: float
    :param gvalue_to_fitness: How to map trait value to fitness
    :type gvalue_to_fitness: fwdpy11.GeneticValueIsTrait
    :param noise: Random effects on trait values
    :type noise: fwdpy11.GeneticValueNoise

    When `gvalue_to_fitness` is `None`, then we are
    modeling multiplicative effects on fitness.

    For a model of fitness, the genetic value is 1, 1+e*h,
    1+`scaling`*e for genotypes AA, Aa, and aa, respectively,
    where `e` and `h` are the effect size and dominance, respectively.

    For a model of a trait (phenotype), meaning `gvalue_to_fitness`
    is not `None`, the values for the three genotypes are 0, e*h,
    and e, respectively.

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    scaling: float
    gvalue_to_fitness: GeneticValueIsTrait = None
    noise: GeneticValueNoise = None
    ndemes: int = 1

    def __attrs_post_init__(self):
        super(Multiplicative, self).__init__(
            self.scaling, self.gvalue_to_fitness, self.noise, self.ndemes
        )


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict_no_recurse
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class GBR(_ll_GBR):
    """
    The "gene-based recessive" trait model described in Thornton et al.
    2013 http://dx.doi.org/10.1371/journal.pgen.1003258 and Sanjak et al. 2017
    http://dx.doi.org/10.1371/journal.pgen.1006573.

    The trait value is the geometric mean of the sum of effect
    sizes on each haplotype.
    It is undefined for the case where these sums are negative.

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param gvalue_to_fitness: How to map trait value to fitness
    :type gvalue_to_fitness: fwdpy11.GeneticValueIsTrait
    :param noise: Random effects on trait values
    :type noise: fwdpy11.GeneticValueNoise

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    gvalue_to_fitness: object
    noise: object = None

    def __attrs_post_init__(self):
        super(GBR, self).__init__(self.gvalue_to_fitness, self.noise)


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict_no_recurse
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class StrictAdditiveMultivariateEffects(_ll_StrictAdditiveMultivariateEffects):
    """
    Multivariate trait values under strictly additive effects.

    Calculate the trait value for a diploid in a
    :class:`fwdpy11.DiploidPopulation` for a multidimensional trait.

    This class is restricted to the case of simple additive effects, meaning
    that any dominance terms associated with mutations are ignored.

    During a simulation, :attr:`fwdpy11.DiploidMetadata.g` is filled with the
    genetic value corresponding to a "focal" trait specified upon
    object construction. This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param ndimensions: Number of trait dimensions
    :type ndimensions: int
    :param focal_trait: Index of the focal trait
    :type focal_trait: int
    :param gvalue_to_fitness: Function mapping trait value to fitness
    :type gvalue_to_fitness: :class:`fwdpy11.GeneticValueToFitnessMap`
    :param noise: Function adding random additive noise to trait value
    :type noise: :class:`fwdpy11.GeneticValueNoise`

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    ndimensions: int
    focal_trait: int
    gvalue_to_fitness: object
    noise: object = None

    def __attrs_post_init__(self):
        super(StrictAdditiveMultivariateEffects, self).__init__(
            self.ndimensions,
            self.focal_trait,
            self.gvalue_to_fitness,
            self.noise
        )



import decimal
import typing

import attr
import demes
import numpy as np

from ..class_decorators import (
    attr_class_pickle_with_super,
    attr_class_to_from_dict,
)

import fwdpy11


def _round_via_decimal(value):
    with decimal.localcontext() as ctx:
        ctx.rounding = decimal.ROUND_HALF_UP
        return int(decimal.Decimal(value).to_integral_value())


@attr.s(repr_ns="fwdpy11")
@attr_class_pickle_with_super
@attr_class_to_from_dict
class ForwardDemesGraph(fwdpy11._fwdpy11._ForwardDemesGraph):
    """
    A forward-in-time representation of a `demes.Graph`

    :param yaml: The string representation of a demes model.
    :type yaml: str
    :param burnin: The number of generations of evolution to occur before the
                   events in the `yaml` take place.
    :type burnin: int
    :param burnin_is_exact: If `False`, `burnin` will be treated as an integer
                            multiple of the sum of ancestor population sizes.
                            If `True`, `burnin` will be used as-is.
    :type burnin_is_exact: bool
    :param round_non_integer_sizes: If `True`, a graph containing non-integer
                            values for epoch start and/or end sizes will
                            have those values rounded to the nearest integer.
    :type round_non_integer_sizes: bool

    The recommended method for construction is
    :meth:`ForwardDemesGraph.from_demes` or
    :meth:`ForwardDemesGraph.tubes`.

    .. versionadded:: 0.20.0
    """
    yaml: str = attr.ib()
    burnin: int = attr.ib()
    burnin_is_exact: int = attr.ib()
    round_non_integer_sizes: bool = attr.ib()
    graph: demes.Graph
    burnin_generation: int

    def __attrs_post_init__(self):
        self.graph = demes.loads(self.yaml)
        Nref = self._get_ancestral_population_size(self.graph)
        assert np.modf(Nref)[0] == 0.0
        if self.burnin_is_exact is True:
            burnin = self.burnin
        else:
            burnin = self.burnin*Nref
        self.burnin_generation = burnin
        super(ForwardDemesGraph, self).__init__(
            self.yaml, burnin, self.round_non_integer_sizes)
        x = self._sum_deme_sizes_at_time_zero()
        assert Nref == x, f"{Nref}, {x}, {self.yaml}"

    def __get_most_ancient_deme_start_time(self, dg: demes.Graph) -> demes.demes.Time:
        return max([d.start_time for d in dg.demes])

    def _get_ancestral_population_size(self, dg: demes.Graph) -> int:
        """
        Need this for the burnin time.

        If there are > 1 demes with the same most ancient start_time,
        then the ancestral size is considered to be the size
        of all those demes (size of ancestral metapopulation).
        """
        oldest_deme_time = self.__get_most_ancient_deme_start_time(dg)

        rv = sum(
            [
                _round_via_decimal(e.start_size)
                for d in dg.demes
                for e in d.epochs
                if e.start_time == oldest_deme_time
            ]
        )
        if rv == 0:
            raise ValueError(
                "could not determinine ancestral metapopulation size")
        return rv

    def number_of_demes(self) -> int:
        return len(self.graph.demes)

    @classmethod
    def from_demes(cls, model: typing.Union[str, demes.Graph], burnin: int, *,
                   burnin_is_exact: typing.Optional[bool] = None,
                   round_non_integer_sizes: typing.Optional[bool] = None):
        """
        Build from a demes graph.

        :param model: A `demes` model.
        :type model: str or demes.Graph
        :param burnin: The number of generations of evolution to occur before
                       the events in the `yaml` take place.
        :type burnin: int
        :param burnin_is_exact: If `False`, `burnin` will be treated as an
                                integer multiple of the sum of ancestor
                                population sizes. If `True`, `burnin` will
                                be used as-is.
        :type burnin_is_exact: bool
        :type round_non_integer_sizes: If `True`, a graph containing
                                non-integer values for epoch start
                                and/or end sizes will have those values
                                rounded to the nearest integer.

        :returns: A forward-time representation of a demes graph
        :rtype: fwdpy11.ForwardDemesGraph
        """
        # NOTE: the logic is messy and we are losing
        # track of "provenance" of the demes input
        if isinstance(model, str) is False:
            yaml = str(model)
        else:
            yaml = model

        if burnin_is_exact is None:
            exact = False
        else:
            exact = burnin_is_exact

        if round_non_integer_sizes is None:
            round_sizes = False
        else:
            round_sizes = round_non_integer_sizes

        return cls(yaml, burnin, exact, round_sizes)

    @classmethod
    def tubes(cls, sizes: typing.List[int], burnin: int, *,
              burnin_is_exact: typing.Optional[bool] = None,
              round_non_integer_sizes: typing.Optional[bool] = None):
        """
        Build a demographic model for one or more demes of constant size.

        :param sizes: Deme sizes
        :type sizes: typing.List[int]
        :param burnin: The number of generations of evolution to simulate.
        :type burnin: int
        :param burnin_is_exact: If `False`, `burnin` will be treated as an
                                integer multiple of the sum of ancestor
                                population sizes. If `True`, `burnin` will
                                be used as-is.
        :type burnin_is_exact: bool
        :type round_non_integer_sizes: If `True`, a graph containing
                                non-integer values for epoch start
                                and/or end sizes will have those values
                                rounded to the nearest integer.

        :returns: A forward-time representation of a demes graph
        :rtype: fwdpy11.ForwardDemesGraph
        """
        builder = demes.Builder()
        for i, j in enumerate(sizes):
            builder.add_deme(f"deme{i}", epochs=[dict(start_size=j)])
        graph = builder.resolve()
        return cls.from_demes(graph, burnin,
                              burnin_is_exact=burnin_is_exact,
                              round_non_integer_sizes=round_non_integer_sizes)

    @property
    def initial_sizes(self) -> typing.List:
        """
        A list of the nonzero (parental) deme sizes at time 0 of the model
        (thinking forwards in time).

        Due to the structure of a demes graph, the indexes of the list
        are the deme integer identifiers.

        This property is useful for setting up a
        :class:`fwdpy11.DiploidPopulation`.
        """
        return [i for i in self._parental_deme_sizes_at_time_zero() if i > 0]

    @property
    def final_generation(self) -> int:
        """
        This is the final offspring generation of the model.
        In other words, a model evolved from parental generation
        zero until time zero of the graph (backwards in time)
        will result in individuals whose birth times equal
        the value of this property.
        """
        return self._model_end_time() - 1

    @property
    def deme_labels(self) -> typing.Dict[int, str]:
        """
        A dictionary mapping integer deme IDs to deme names
        (strings).
        """
        return {i: j.name for i, j in enumerate(self.graph.demes)}

    @property
    def demes_at_final_generation(self) -> typing.List[int]:
        """
        List of integer ids of extant demes in the final generation
        of the demes graph.
        """
        epoch_end_times = []
        for deme in self.graph.demes:
            for epoch in deme.epochs:
                epoch_end_times.append(epoch.end_time)
        min_end_time = min(epoch_end_times)
        rv = []
        for deme in self.deme_labels.keys():
            for epoch in self.graph.demes[deme].epochs:
                if epoch.end_time == min_end_time:
                    rv.append(deme)
        return rv



#
# Copyright (C) 2017-2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest

import numpy as np

import fwdpy11


def build_model(gvalue_to_fitness, noise, simlen, sizes):
    yaml = f"""
time_units: generations
demes:
 - name: A
   epochs:
    - start_size: {sizes[0]}
 - name: B
   epochs:
    - start_size: {sizes[1]}
migrations:
 - demes: [A, B]
   rate: 0.1
"""
    pdict = {
        "nregions": [],
        "sregions": [
            fwdpy11.mvDES(
                fwdpy11.MultivariateGaussianEffects(
                    0, 1, 1, h=0.25, cov_matrix=np.identity(2)
                ),
                np.zeros(2),
            )
        ],
        "recregions": [fwdpy11.PoissonInterval(0, 1, 0.5)],
        "rates": (0, 1e-2, None),
        "demography": fwdpy11.ForwardDemesGraph.from_demes(yaml, burnin=simlen,
                                                           burnin_is_exact=True),
        "simlen": simlen,
        "gvalue": fwdpy11.Additive(
            ndemes=2, scaling=2, gvalue_to_fitness=gvalue_to_fitness, noise=noise,
        ),
    }
    return pdict


def build_single_trait_model(gvalue, simlen):
    pdict = {
        "nregions": [],
        "sregions": [fwdpy11.GaussianS(0.0, 1.0, 1.0, 0.1)],
        "recregions": [fwdpy11.PoissonInterval(0, 1, 0.5)],
        "rates": (0, 1e-3, None),
        "simlen": simlen,
        "gvalue": gvalue,
    }
    return pdict


class TestCustomGeneticValueisTrait(unittest.TestCase):
    def test_run(self):

        import pygss

        GSS = pygss.PyGSS(opt=0.0, VS=1.0)
        pdict = build_model(GSS, None, 100, [1000, 1000])

        params = fwdpy11.ModelParams(**pdict)
        pop = fwdpy11.DiploidPopulation([1000, 1000], 1.0)

        rng = fwdpy11.GSLrng(1010)
        fwdpy11.evolvets(rng, pop, params, 100, suppress_table_indexing=True)

        self.assertEqual(pop.generation, params.simlen)
        md = np.array(pop.diploid_metadata)

        pdict["gvalue"] = fwdpy11.Additive(
            # ndemes=2, scaling=2, gvalue_to_fitness=
            ndemes=2,
            scaling=2,
            gvalue_to_fitness=fwdpy11.GaussianStabilizingSelection.single_trait([
                fwdpy11.Optimum(optimum=0.0, VS=1.0)]),
        )
        params = fwdpy11.ModelParams(**pdict)
        pop2 = fwdpy11.DiploidPopulation([1000, 1000], 1.0)
        rng = fwdpy11.GSLrng(1010)
        fwdpy11.evolvets(rng, pop2, params, 100, suppress_table_indexing=True)
        md2 = np.array(pop.diploid_metadata)
        self.assertAlmostEqual(md["g"].mean(), md2["g"].mean())

    def test_run_stateful(self):

        import pygss

        GSS = pygss.PyGSSRandomOptimum(opt=0.0, VS=1.0)
        pdict = build_model(GSS, None, 5, [1000, 1000])
        params = fwdpy11.ModelParams(**pdict)
        pop = fwdpy11.DiploidPopulation([1000, 1000], 1.0)

        rng = fwdpy11.GSLrng(1010)
        fwdpy11.evolvets(rng, pop, params, 100, suppress_table_indexing=True)
        self.assertEqual(
            len(params.gvalue.gvalue_to_fitness.optima), pop.generation + 1
        )


class TestCustomGeneticValueNoise(unittest.TestCase):
    def test_run(self):

        import pynoise

        GSS = fwdpy11.fwdpy11.GaussianStabilizingSelection.single_trait(
            [fwdpy11.Optimum(optimum=0.0, VS=1.0, when=0)])
        Noise = pynoise.PyNoise()
        pdict = build_model(GSS, Noise, 5, [1000, 1000])
        params = fwdpy11.ModelParams(**pdict)
        pop = fwdpy11.DiploidPopulation([1000, 1000], 1.0)

        rng = fwdpy11.GSLrng(1010)
        fwdpy11.evolvets(rng, pop, params, 100, suppress_table_indexing=True)

        self.assertEqual(pop.generation, params.simlen)
        md = np.array(pop.diploid_metadata)
        self.assertTrue(len(np.where(md["e"] > 0)[0]) > 0)


class TestCustomPyGeneticValue(unittest.TestCase):
    @classmethod
    def setUp(self):
        import pygss
        import pyadditive
        import pynoise

        PyA = pyadditive.PyAdditive(
            pygss.PyGSS(opt=0.0, VS=1.0), noise=pynoise.PyNoise()
        )

        self.pdict_PyA = build_single_trait_model(PyA, 100)

    def test_run(self):
        pop_PyA = fwdpy11.DiploidPopulation(1000, 1.0)
        self.pdict_PyA["demography"] = \
            fwdpy11.ForwardDemesGraph.tubes(pop_PyA.deme_sizes()[1],
                                            burnin=self.pdict_PyA["simlen"],
                                            burnin_is_exact=True)
        mparams_PyA = fwdpy11.ModelParams(**self.pdict_PyA)
        rng_PyA = fwdpy11.GSLrng(42 * 666)

        fwdpy11.evolvets(rng_PyA, pop_PyA, mparams_PyA, 100)
        if len(pop_PyA.tables.mutations) == 0:
            self.fail("simulation ended with no mutations")
        md = np.array(pop_PyA.diploid_metadata, copy=False)
        self.assertTrue(md["g"].var() > 0.0)
        self.assertTrue(md["e"].var() > 0.0)


class TestCustomPyGeneticValueOverloadGeneticValueToFitness(unittest.TestCase):
    def build_model(self, gvalue, simlen):
        pdict = {
            "nregions": [],
            "sregions": [fwdpy11.GaussianS(beg=0.0, end=1.0, weight=1.0, sd=0.1)],
            "recregions": [fwdpy11.PoissonInterval(0, 1, 0.5)],
            "rates": (0, 1e-2, None),
            "demography": None,
            "simlen": simlen,
            "gvalue": gvalue,
        }
        return pdict

    def test_run(self):

        import pyadditivegss

        pygv = pyadditivegss.PyAdditiveGSS(opt=0.0, VS=1.0)
        pdict = self.build_model(pygv, 100)
        pop = fwdpy11.DiploidPopulation(1000, 1.0)
        pdict["demography"] = \
            fwdpy11.ForwardDemesGraph.tubes(pop.deme_sizes()[1],
                                            burnin=pdict["simlen"],
                                            burnin_is_exact=True)

        params = fwdpy11.ModelParams(**pdict)

        rng = fwdpy11.GSLrng(1010)
        fwdpy11.evolvets(rng, pop, params, 100, suppress_table_indexing=True)

        self.assertEqual(pop.generation, params.simlen)
        md = np.array(pop.diploid_metadata)

        pdict["gvalue"] = fwdpy11.Additive(
            2.0, fwdpy11.GaussianStabilizingSelection.single_trait([fwdpy11.Optimum(optimum=0.0, VS=1.0, when=0)]))
        params = fwdpy11.ModelParams(**pdict)
        pop2 = fwdpy11.DiploidPopulation(1000, 1.0)
        rng = fwdpy11.GSLrng(1010)
        fwdpy11.evolvets(rng, pop2, params, 100, suppress_table_indexing=True)
        md2 = np.array(pop.diploid_metadata)
        self.assertAlmostEqual(md["g"].mean(), md2["g"].mean())


if __name__ == "__main__":
    unittest.main()



def setup_mako(cfg):
    # import fwdpy11 so we can find its C++ headers
    import fwdpy11 as fp11
    import subprocess
    import sys

    gsl_prefix = (
        subprocess.check_output(["gsl-config", "--prefix"]).strip().decode("utf-8")
    )
    # add fwdpy11 header locations to the include path
    cfg["include_dirs"].extend(
        [fp11.get_includes(), fp11.get_fwdpp_includes(), gsl_prefix + str("/include")]
    )
    gsl_libs = subprocess.check_output(["gsl-config", "--libs"]).strip().decode("utf-8")
    gsl_libs = gsl_libs.split(" ")
    gsl_runtime_libs = [i.replace("-l", "") for i in gsl_libs[1:]]
    cfg["libraries"].extend(gsl_runtime_libs)
    cfg["extra_link_args"].extend([gsl_libs[0]])
    # On OS X using clang, there is more work to do.  Using gcc on OS X
    # gets rid of these requirements. The specifics sadly depend on how
    # you initially built fwdpy11, and what is below assumes you used
    # the provided setup.py + OS X + clang:
    if sys.platform == "darwin":
        cfg["compiler_args"].extend(["-stdlib=libc++", "-mmacosx-version-min=10.7"])
        cfg["linker_args"] = ["-stdlib=libc++", "-mmacosx-version-min=10.7"]



# Unit tests for fwdpy11.regions

import pickle
import unittest

import numpy as np

import fwdpy11

# Choose global values that differ from default
# values in Python class constructors
BEG, END, WEIGHT, DOM, LABEL, COUPLED = 0.0, 1.0, 0.5, 0.5, 63, False


class testRegion(unittest.TestCase):
    """
    These tests all apply to Sregions b/c an
    Sregion contains a Region
    """

    def testBadBoundaries(self):
        with self.assertRaises(ValueError):
            fwdpy11.Region(np.nan, 1, 1.0)
        with self.assertRaises(ValueError):
            fwdpy11.Region(0, np.nan, 1.0)
        with self.assertRaises(ValueError):
            fwdpy11.Region(0, 0, 1.0)
        with self.assertRaises(ValueError):
            fwdpy11.Region(0, -1, 1.0)

    def testBadWeight(self):
        with self.assertRaises(ValueError):
            fwdpy11.Region(0, 1, np.nan)
        with self.assertRaises(ValueError):
            fwdpy11.Region(0, 1, -1e-3)


class test_PickleRegion(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.r = fwdpy11.Region(BEG, END, WEIGHT, COUPLED, LABEL)

    def test_pickling(self):
        p = pickle.dumps(self.r, -1)
        up = pickle.loads(p)
        self.assertEqual(self.r.beg, up.beg)
        self.assertEqual(self.r.end, up.end)
        self.assertEqual(self.r.weight, up.weight)
        self.assertEqual(self.r.coupled, up.coupled)
        self.assertEqual(self.r.label, up.label)


class testExpS(unittest.TestCase):
    # NOTE: this tests applies to all Sregions
    # b/c the scaling param is part of the base class
    def testBadScaling(self):
        with self.assertRaises(ValueError):
            fwdpy11.ExpS(0, 1, 1.0, -0.1, scaling=np.nan)

    def testBadMean(self):
        with self.assertRaises(ValueError):
            fwdpy11.ExpS(0, 1, 1, np.nan)

    def testBadDominance(self):
        with self.assertRaises(ValueError):
            fwdpy11.ExpS(0, 1, 1, -1.0, h=np.nan)


class test_PickleExpS(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.r = fwdpy11.ExpS(BEG, END, WEIGHT, -0.2, DOM, COUPLED, LABEL)

    def test_pickling(self):
        p = pickle.dumps(self.r, -1)
        up = pickle.loads(p)
        self.assertEqual(self.r.beg, up.beg)
        self.assertEqual(self.r.end, up.end)
        self.assertEqual(self.r.weight, up.weight)
        self.assertEqual(self.r.coupled, up.coupled)
        self.assertEqual(self.r.label, up.label)
        self.assertEqual(up.mean, -0.2)
        self.assertEqual(up.h, DOM)


class testGammaS(unittest.TestCase):
    def testBadMean(self):
        with self.assertRaises(ValueError):
            fwdpy11.GammaS(0, 1, 1, mean=np.nan, shape_parameter=1.0)

    def testBadShape(self):
        with self.assertRaises(ValueError):
            fwdpy11.GammaS(0, 1, 1, mean=1.0, shape_parameter=np.nan)

    def testBadDominance(self):
        with self.assertRaises(ValueError):
            fwdpy11.GammaS(0, 1, 1, mean=1.0, shape_parameter=1.0, h=np.nan)


class test_PickleGammaS(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.r = fwdpy11.GammaS(BEG, END, WEIGHT, -0.2, 0.5, DOM, COUPLED, LABEL)

    def test_pickling(self):
        p = pickle.dumps(self.r, -1)
        up = pickle.loads(p)
        self.assertEqual(self.r.beg, up.beg)
        self.assertEqual(self.r.end, up.end)
        self.assertEqual(self.r.weight, up.weight)
        self.assertEqual(self.r.coupled, up.coupled)
        self.assertEqual(self.r.label, up.label)
        self.assertEqual(up.mean, -0.2)
        self.assertEqual(up.shape_parameter, 0.5)
        self.assertEqual(up.h, DOM)


class testUniformS(unittest.TestCase):
    def testBadRange(self):
        with self.assertRaises(ValueError):
            fwdpy11.UniformS(0, 1, 1, 0, 0)
        with self.assertRaises(ValueError):
            fwdpy11.UniformS(0, 1, 1, np.nan, 1)
        with self.assertRaises(ValueError):
            fwdpy11.UniformS(0, 1, 1, 1, np.nan)

    def testBadDominance(self):
        with self.assertRaises(ValueError):
            fwdpy11.UniformS(0, 1, 1, lo=1.0, hi=1.0, h=np.nan)


class test_PickleUniformS(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.r = fwdpy11.UniformS(BEG, END, WEIGHT, -0.2, 0.5, DOM, COUPLED, LABEL)

    def test_pickling(self):
        p = pickle.dumps(self.r, -1)
        up = pickle.loads(p)
        self.assertEqual(self.r.beg, up.beg)
        self.assertEqual(self.r.end, up.end)
        self.assertEqual(self.r.weight, up.weight)
        self.assertEqual(self.r.coupled, up.coupled)
        self.assertEqual(self.r.label, up.label)
        self.assertEqual(up.lo, -0.2)
        self.assertEqual(up.hi, 0.5)
        self.assertEqual(up.h, DOM)


class testGaussianS(unittest.TestCase):
    def testBadStdDev(self):
        with self.assertRaises(ValueError):
            fwdpy11.GaussianS(0, 1, 1, np.nan)
        with self.assertRaises(ValueError):
            fwdpy11.GaussianS(0, 1, 1, 0)
        with self.assertRaises(ValueError):
            fwdpy11.GaussianS(0, 1, 1, -1e-3)

    def testBadDominance(self):
        with self.assertRaises(ValueError):
            fwdpy11.GaussianS(0, 1, 1, 1e-3, h=np.nan)


class test_PickleGaussianS(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.r = fwdpy11.GaussianS(BEG, END, WEIGHT, 0.5, DOM, COUPLED, LABEL)

    def test_pickling(self):
        p = pickle.dumps(self.r, -1)
        up = pickle.loads(p)
        self.assertEqual(self.r.beg, up.beg)
        self.assertEqual(self.r.end, up.end)
        self.assertEqual(self.r.weight, up.weight)
        self.assertEqual(self.r.coupled, up.coupled)
        self.assertEqual(self.r.label, up.label)
        self.assertEqual(up.sd, 0.5)
        self.assertEqual(up.h, DOM)


class testConstantS(unittest.TestCase):
    def testBadValue(self):
        with self.assertRaises(ValueError):
            fwdpy11.ConstantS(0, 1, 1, np.nan)

    def testBadDominance(self):
        with self.assertRaises(ValueError):
            fwdpy11.ConstantS(0, 1, 1, 1, np.nan)


class test_PickleConstantS(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.r = fwdpy11.ConstantS(BEG, END, WEIGHT, 0.5, DOM, COUPLED, LABEL)

    def test_pickling(self):
        p = pickle.dumps(self.r, -1)
        up = pickle.loads(p)
        self.assertEqual(self.r.beg, up.beg)
        self.assertEqual(self.r.end, up.end)
        self.assertEqual(self.r.weight, up.weight)
        self.assertEqual(self.r.coupled, up.coupled)
        self.assertEqual(self.r.label, up.label)
        self.assertEqual(up.s, 0.5)
        self.assertEqual(up.h, DOM)


class Test_mvDES(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.r = fwdpy11.mvDES(
            [fwdpy11.ExpS(0, 1, 1, 0.1), fwdpy11.ExpS(0, 1, 1, -0.1)],
            np.zeros(2),
            np.identity(2),
        )

    def test_pickling(self):
        p = pickle.dumps(self.r, -1)
        up = pickle.loads(p)
        self.assertTrue(np.array_equal(up.means, self.r.means))
        self.assertTrue(np.array_equal(up.matrix, np.identity(2)))

    def test_bad_init_1(self):
        with self.assertRaises(ValueError):
            fwdpy11.mvDES(
                [fwdpy11.ExpS(0, 1, 1, 0.1), fwdpy11.ExpS(0, 1, 1, -0.1)],
                np.zeros(3),
                np.identity(2),
            )

    def test_mvLogNormalS(self):
        try:
            mvln = fwdpy11.LogNormalS.mv(0, 1, 1)
            fwdpy11.mvDES(mvln, np.zeros(5), np.identity(5))
        except:  # NOQA
            self.fail("unexpected exception")

    def test_mvLogNormalS_pickling(self):
        mvln = fwdpy11.LogNormalS.mv(0, 1, 1)
        des = fwdpy11.mvDES(mvln, np.zeros(5), np.identity(5))
        p = pickle.dumps(des, -1)
        pickle.loads(p)

    def test_mvLogNormalS_bad_init(self):
        mvln = fwdpy11.LogNormalS.mv(0, 1, 1)
        with self.assertRaises(ValueError):
            fwdpy11.mvDES([mvln], np.zeros(5), np.identity(5))

    def test_MultivariateGaussian(self):
        try:
            mvg = fwdpy11.MultivariateGaussianEffects(0, 1, 1, np.identity(4))
            fwdpy11.mvDES(mvg, np.zeros(4))
        except:  # NOQA
            self.fail("unexpected exception")

    def test_MultivariateGaussian_pickling(self):
        mvg = fwdpy11.MultivariateGaussianEffects(0, 1, 1, np.identity(4))
        des = fwdpy11.mvDES(mvg, np.zeros(4))
        p = pickle.dumps(des, -1)
        pickle.loads(p)

    def test_MultivariateGaussian_bad_init(self):
        mvg = fwdpy11.MultivariateGaussianEffects(0, 1, 1, np.identity(4))
        with self.assertRaises(ValueError):
            fwdpy11.mvDES([mvg], np.zeros(4), np.identity(4))


class testMutationRegions(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.nregions = [fwdpy11.Region(0, 1, 1)]
        self.sregions = [
            fwdpy11.ConstantS(0, 1, 1, -0.1, 0.25),
            fwdpy11.ExpS(0, 1, 0.001, 0.01, 1.0),
        ]

    def test_create(self):
        mr = fwdpy11.MutationRegions.create(0.5, self.nregions, self.sregions)  # NOQA

    def test_shape(self):
        self.assertEqual(self.sregions[0].shape, (1,))


class testMultivariateGaussianEffects(unittest.TestCase):
    def testConstruction(self):
        try:
            fwdpy11.MultivariateGaussianEffects(0, 1, 1, np.identity(2))
        except Exception:
            self.fail("Unexpected exception during object construction")

    def test_shape(self):
        mv = fwdpy11.MultivariateGaussianEffects(0, 1, 1, np.identity(2))
        self.assertEqual(mv.shape, (2,))

    def testInvalidMatrixShape(self):
        # Input matrix has wrong shape:
        m = np.array([-1.0] * 4)
        with self.assertRaises(ValueError):
            fwdpy11.MultivariateGaussianEffects(0, 1, 1, m)

    def testInvalidMatrixShape2(self):
        m = np.array([-1.0] * 6).reshape((3, 2))
        with self.assertRaises(ValueError):
            fwdpy11.MultivariateGaussianEffects(0, 1, 1, m)

    def testInvalidMatrix(self):
        # This is not a positive-definite covariance
        # matrix:
        m = np.array([-1.0] * 4).reshape((2, 2))
        with self.assertRaises(ValueError):
            fwdpy11.MultivariateGaussianEffects(0, 1, 1, m)

    def testMatrixWithNAN(self):
        m = np.identity(4).reshape((4, 4))
        m[0, 1] = np.nan
        with self.assertRaises(ValueError):
            fwdpy11.MultivariateGaussianEffects(0, 1, 1, m)


class test_PickleMultivariateGaussianEffects(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.r = fwdpy11.MultivariateGaussianEffects(
            BEG, END, WEIGHT, np.identity(2), -0.3, DOM, COUPLED, LABEL
        )

    def test_pickling(self):
        p = pickle.dumps(self.r, -1)
        up = pickle.loads(p)
        self.assertEqual(self.r, up)


class testRecombinationRegions(unittest.TestCase):
    def test_create(self):
        rr = fwdpy11.RecombinationRegions(
            1e-3, [fwdpy11.Region(0, 1, 1), fwdpy11.Region(1, 2, 1)]
        )
        self.assertEqual(len(rr.weights), 2)

    def test_bad_input(self):
        with self.assertRaises(TypeError):
            fwdpy11.RecombinationRegions(
                1e-3, [fwdpy11.ExpS(0, 1, 1, -0.2), fwdpy11.GaussianS(1, 2, 1, 0.25)]
            )


class TestLogNormalS(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.lns = fwdpy11.LogNormalS(0, 1, 1, 0.1, 2)

    def test_create(self):
        self.assertEqual(self.lns.beg, 0)
        self.assertEqual(self.lns.end, 1)
        self.assertEqual(self.lns.weight, 1)
        self.assertEqual(self.lns.zeta, 0.1)
        self.assertEqual(self.lns.sigma, 2)

    def test_pickling(self):
        p = pickle.dumps(self.lns, -1)
        up = pickle.loads(p)
        self.assertEqual(up.beg, 0)
        self.assertEqual(up.end, 1)
        self.assertEqual(up.weight, 1)
        self.assertEqual(up.zeta, 0.1)
        self.assertEqual(up.sigma, 2)


class TestMultivariateLogNormalS(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.lns = fwdpy11.LogNormalS.mv(0, 1, 1, scaling=3.0, coupled=False)

    def test_create(self):
        self.assertEqual(self.lns.beg, 0)
        self.assertEqual(self.lns.end, 1)
        self.assertEqual(self.lns.weight, 1)
        self.assertEqual(self.lns.coupled, False)
        self.assertTrue(self.lns.zeta is None)
        self.assertTrue(self.lns.sigma is None)

    def test_pickling(self):
        p = pickle.dumps(self.lns, -1)
        up = pickle.loads(p)
        self.assertEqual(up.beg, 0)
        self.assertEqual(up.end, 1)
        self.assertEqual(up.weight, 1)
        self.assertEqual(up.coupled, False)
        self.assertEqual(up.scaling, 3.0)
        self.assertTrue(up.zeta is None)
        self.assertTrue(up.sigma is None)


class TestDiscreteDESD(unittest.TestCase):
    def test_init(self):
        jdist = [(0.1, 1, 1), (0.2, 0.5, 1)]
        jd = fwdpy11.DiscreteDESD(0, 1, 2, jdist)

        self.assertEqual(jd.beg, 0)
        self.assertEqual(jd.end, 1)
        self.assertEqual(jd.weight, 2)
        self.assertEqual(jd.joint_dist, jdist)

    def test_pickle(self):
        jdist = [(0.1, 1, 1), (0.2, 0.5, 1)]
        jd = fwdpy11.DiscreteDESD(0, 1, 2, jdist)

        pp = pickle.dumps(jd, -1)
        up = pickle.loads(pp)

        self.assertEqual(jd.beg, up.beg)
        self.assertEqual(jd.end, up.end)
        self.assertEqual(jd.weight, up.weight)
        self.assertEqual(jd.coupled, up.coupled)
        self.assertEqual(jd.scaling, up.scaling)
        self.assertEqual(jd.joint_dist, up.joint_dist)

    def test_bad_init_1(self):
        jdist = [(0.1, 1, np.nan), (0.2, 0.5, 1)]
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDESD(0, 1, 2, jdist)

    def test_bad_init_2(self):
        jdist = [(0.1, 1, -1), (0.2, 0.5, 1)]
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDESD(0, 1, 2, jdist)

    def test_bad_init_3(self):
        jdist = [(0.1, np.nan, 1), (0.2, 0.5, 1)]
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDESD(0, 1, 2, jdist)

    def test_bad_init_4(self):
        jdist = [(np.nan, 1, 1), (0.2, 0.5, 1)]
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDESD(0, 1, 2, jdist)


def test_ConstantS_weight_display():
    c = fwdpy11.ConstantS(0, 10, 1.0, s=0.5, coupled=True)
    sc = str(c)
    print(sc)
    assert "weight=10" in sc
    pc = pickle.dumps(c)
    up = pickle.loads(pc)
    assert "weight=10" in str(up)


if __name__ == "__main__":
    unittest.main()



import pickle

import pytest

import fwdpy11


@pytest.mark.parametrize("discrete", [True, False])
def test_poisson_interval(discrete):
    pi = fwdpy11.PoissonInterval(0, 2, 1e-3, discrete=discrete)
    p = pickle.dumps(pi)
    up = pickle.loads(p)
    assert up.beg == pi.beg
    assert up.end == pi.end
    assert up.mean == pi.mean
    assert up.discrete == pi.discrete
    assert up.discrete == discrete

    if discrete is True:
        with pytest.raises(ValueError):
            fwdpy11.PoissonInterval(0, 1, 1e-3, discrete=discrete)
        with pytest.raises(TypeError):
            fwdpy11.PoissonInterval(0.0, 2, 1e-3, discrete=discrete)
        with pytest.raises(TypeError):
            fwdpy11.PoissonInterval(0, 2.0, 1e-3, discrete=discrete)


@pytest.mark.parametrize("discrete", [True, False])
def test_binomial_interval(discrete):
    pi = fwdpy11.BinomialInterval(0, 2, 1e-3, discrete=discrete)
    p = pickle.dumps(pi)
    up = pickle.loads(p)
    assert up.beg == pi.beg
    assert up.end == pi.end
    assert up.probability == pi.probability
    assert up.discrete == pi.discrete
    assert up.discrete == discrete

    if discrete is True:
        with pytest.raises(ValueError):
            fwdpy11.BinomialInterval(0, 1, 1e-3, discrete=discrete)
        with pytest.raises(TypeError):
            fwdpy11.BinomialInterval(0.0, 1, 1e-3, discrete=discrete)
        with pytest.raises(TypeError):
            fwdpy11.BinomialInterval(0, 1.0, 1e-3, discrete=discrete)


@pytest.mark.parametrize("discrete", [True, False])
def test_poisson_point(discrete):
    pi = fwdpy11.PoissonPoint(0, 1e-3, discrete=discrete)
    p = pickle.dumps(pi)
    up = pickle.loads(p)
    assert up.position == pi.position
    assert up.mean == pi.mean
    assert up.discrete == pi.discrete
    assert up.discrete == discrete

    if discrete is True:
        with pytest.raises(TypeError):
            fwdpy11.PoissonPoint(0.0, 1e-3, discrete=discrete)


@pytest.mark.parametrize("discrete", [True, False])
def test_binomial_point(discrete):
    pi = fwdpy11.BinomialPoint(0, 1e-3, discrete=discrete)
    p = pickle.dumps(pi)
    up = pickle.loads(p)
    assert up.position == pi.position
    assert up.probability == pi.probability
    assert up.discrete == pi.discrete
    assert up.discrete == discrete

    if discrete is True:
        with pytest.raises(TypeError):
            fwdpy11.BinomialPoint(0.0, 1e-3, discrete=discrete)


@pytest.mark.parametrize("discrete", [True, False])
def test_fixed_crossovers(discrete):
    pi = fwdpy11.FixedCrossovers(0, 2, 3, discrete=discrete)
    p = pickle.dumps(pi)
    up = pickle.loads(p)
    assert up.beg == pi.beg
    assert up.end == pi.end
    assert up.num_xovers == pi.num_xovers
    assert up.discrete == pi.discrete
    assert up.discrete == discrete

    if discrete is True:
        with pytest.raises(ValueError):
            fwdpy11.FixedCrossovers(0, 1, 3, discrete=discrete)
        with pytest.raises(TypeError):
            fwdpy11.FixedCrossovers(0.0, 1, 3, discrete=discrete)
        with pytest.raises(TypeError):
            fwdpy11.FixedCrossovers(0, 1.0, 3, discrete=discrete)



#
# Copyright (C) 2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import pickle

import fwdpy11
import numpy as np
import pytest


@pytest.mark.parametrize(
    "t", [(fwdpy11.FixedDominance, "h"), (fwdpy11.ExponentialDominance, "m")]
)
@pytest.mark.parametrize("x", [1.0, -0.25])
def test_single_param_construction(t, x):
    _ = t[0](x)
    kwargs = {t[1]: x}
    _ = t[0](**kwargs)


@pytest.mark.parametrize(
    "t", [(fwdpy11.FixedDominance, "h"), (fwdpy11.ExponentialDominance, "m")]
)
@pytest.mark.parametrize("x", [np.nan, np.inf])
def test_single_param_bad_construction(t, x):
    with pytest.raises(ValueError):
        _ = t[0](x)
    with pytest.raises(ValueError):
        kwargs = {t[1]: x}
        _ = t[0](**kwargs)


@pytest.mark.parametrize("x", [(0.0, 1.0), (-0.25, 1.0), (-1.0, 0.0), (-1.0, -1 / 3)])
def test_UniformDominance_construction(x):
    _ = fwdpy11.UniformDominance(lo=x[0], hi=x[1])
    _ = fwdpy11.UniformDominance(x[0], x[1])


@pytest.mark.parametrize(
    "x", [(np.inf, 1.0), (np.nan, 1.0), (1.0, np.inf), (np.nan, 1.0), (1.0, 0.0)]
)
def test_UniformDominance_bad_construction(x):
    with pytest.raises(ValueError):
        _ = fwdpy11.UniformDominance(lo=x[0], hi=x[1])


@pytest.mark.parametrize("k", [1e-3, 1.0, 10.0])
def test_LargeEffectExponentiallyRecessive_construction(k):
    _ = fwdpy11.LargeEffectExponentiallyRecessive(k)
    _ = fwdpy11.LargeEffectExponentiallyRecessive(k=k)


@pytest.mark.parametrize("k", [-1e-3, np.nan, np.inf])
def test_LargeEffectExponentiallyRecessive_bad_construction(k):
    with pytest.raises(ValueError):
        _ = fwdpy11.LargeEffectExponentiallyRecessive(k)
    with pytest.raises(ValueError):
        _ = fwdpy11.LargeEffectExponentiallyRecessive(k=k)


@pytest.mark.parametrize(
    "x",
    [
        fwdpy11.FixedDominance(h=0.667),
        fwdpy11.ExponentialDominance(m=-0.23),
        fwdpy11.LargeEffectExponentiallyRecessive(k=1.0),
        fwdpy11.LargeEffectExponentiallyRecessive(k=1.0, scaling=1.),
        fwdpy11.UniformDominance(lo=-2 / 3, hi=2 / 3),
    ],
)
def test_pickling(x):
    p = pickle.dumps(x, -1)
    up = pickle.loads(p)
    assert up == x


@pytest.mark.parametrize(
    "des",
    [
        (fwdpy11.ExpS, {"mean": -0.2}),
        (fwdpy11.ConstantS, {"s": -0.2}),
        (fwdpy11.GammaS, {"mean": 1, "shape_parameter": 1}),
        (fwdpy11.GaussianS, {"sd": 1}),
        (fwdpy11.LogNormalS, {"zeta": 1, "sigma": 1}),
        (fwdpy11.UniformS, {"lo": -0.1, "hi": -1e-3}),
        (fwdpy11.MultivariateGaussianEffects, {"cov_matrix": np.identity(2)}),
    ],
)
@pytest.mark.parametrize(
    "h",
    [
        fwdpy11.FixedDominance(h=0.667),
        fwdpy11.ExponentialDominance(m=-0.23),
        fwdpy11.LargeEffectExponentiallyRecessive(k=1.0),
        fwdpy11.LargeEffectExponentiallyRecessive(k=1.0, scaling=0.5),
        fwdpy11.UniformDominance(lo=-2 / 3, hi=2 / 3),
    ],
)
def test_build_DES_objects(des, h):
    x = des[0](beg=0, end=1, weight=1, h=h, **des[1])
    p = pickle.dumps(x)
    up = pickle.loads(p)
    assert x == up



def _make_cloner(cls, base, *args):
    def clone(self):
        # create a new object without initializing it
        cloned = cls.__new__(cls)
        # clone C++ state
        base.__init__(cloned, *args)
        # clone Python state
        cloned.__dict__.update(self.__dict__)
        return cloned

    cls.clone = clone
    return cls


class genetic_value_is_trait_default_clone(object):
    def __init__(self, ndim=1):
        self.ndim = ndim

    def __call__(self, cls):
        from fwdpy11 import GeneticValueIsTrait

        return _make_cloner(cls, GeneticValueIsTrait, self.ndim)


def genetic_value_noise_default_clone(cls):
    from fwdpy11 import GeneticValueNoise

    return _make_cloner(cls, GeneticValueNoise)
    return cls


def default_update(cls):
    def update(self, pop):
        pass

    cls.update = update
    return cls



#
# Copyright (C) 2017-2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

from typing import List, Optional, Union

import numpy as np

from .._fwdpy11 import MutationRecord, Site, ll_VariantIterator
from .table_collection import TableCollection


class VariantIterator(ll_VariantIterator):
    """
    An iterable class for traversing genotypes in a tree sequence.

    To initialize:

    :param tables: The table collection
    :type tables: fwdpy11.TableCollection
    :param samples: Samples list
    :type samples: list or numpy.ndarray

    The following are `kwarg`-only:

    :param begin: (0.0) First position, inclusive.
    :type begin: float
    :param end: (None) Last position, exclusive.
    :type end: float
    :param include_neutral_variants: (True) Include neutral variants during traversal
    :type include_neutral_variants: bool
    :param include_selected_variants: (True) Include selected variants during traversal
    :type include_selected_variants: bool

    .. versionchanged:: 0.4.1

         Add begin, end options as floats

    .. versionchanged:: 0.4.2

         Add include_neutral_variants and include_selected_variants

    .. versionchanged:: 0.5.0

         No longer requires a :class:`fwdpy11.MutationVector`.

    .. versionchanged:: 0.14.0

         Initialization now requires a TableCollection.
         Some initialization parmeters are now keyword-only.

    """

    def __init__(
        self,
        tables: TableCollection,
        samples: Union[List, np.ndarray],
        *,
        begin: float = 0.0,
        end: Optional[float] = None,
        include_neutral_variants: bool = True,
        include_selected_variants: bool = True
    ):
        if end is not None:
            _end = end
        else:
            _end = float(np.finfo(np.float64).max)

        super(VariantIterator, self).__init__(
            tables,
            samples,
            begin,
            _end,
            include_neutral_variants,
            include_selected_variants,
        )

    @property
    def genotypes(self) -> np.ndarray:
        """Genotype array.  Index order is same as sample input"""
        return self._genotypes

    @property
    def position(self) -> float:
        """Current mutation position"""
        return self._position

    @property
    def records(self) -> List[MutationRecord]:
        """:class:`fwdpy11.MutationRecord` objects for the current Site."""
        return self._records

    @property
    def site(self) -> Site:
        """The current :class:`fwdpy11.Site`"""
        return self._site



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import pickle
import unittest

import fwdpy11


class testSingleEffectMutation(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.m = fwdpy11.Mutation(1.0, -1.0, 0.25, 0, 13)

    def testConstruct(self):
        self.assertEqual(self.m.pos, 1.0)
        self.assertEqual(self.m.s, -1.0)
        self.assertEqual(self.m.h, 0.25)
        self.assertEqual(self.m.g, 0)
        self.assertEqual(self.m.label, 13)
        self.assertEqual(len(self.m.esizes), 0)
        self.assertEqual(len(self.m.heffects), 0)

    def testPickle(self):
        p = pickle.dumps(self.m)
        up = pickle.loads(p)
        self.assertEqual(up, self.m)


class testMultiEffectMutation(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.m = fwdpy11.Mutation(1.0, -1.0, 0.25, 0, [1.0, 2.0], [-1.0, 1.0 / 3.0], 13)

    def testConstruct(self):
        self.assertEqual(self.m.pos, 1.0)
        self.assertEqual(self.m.s, -1.0)
        self.assertEqual(self.m.h, 0.25)
        self.assertEqual(self.m.g, 0)
        self.assertEqual(self.m.label, 13)
        self.assertEqual(len(self.m.esizes), 2)
        self.assertEqual(len(self.m.heffects), 2)

    def testPickle(self):
        p = pickle.dumps(self.m)
        up = pickle.loads(p)
        self.assertEqual(up, self.m)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017-2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
"""
    isort:skip_file
"""
from .population_mixin import PopulationMixin  # NOQA
from .table_collection import TableCollection  # NOQA
from .tree_iterator import TreeIterator  # NOQA
from .variant_iterator import VariantIterator  # NOQA
from .data_matrix import DataMatrix  # NOQA
from .data_matrix_iterator import DataMatrixIterator  # NOQA
from .diploid_population import DiploidPopulation  # NOQA
from .model_params import ModelParams  # NOQA
from .new_mutation_data import NewMutationData  # NOQA
from .demographic_model_citation import DemographicModelCitation  # NOQA
from .demographic_model_details import DemographicModelDetails  # NOQA
from .forward_demes_graph import ForwardDemesGraph  # NOQA



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import attr


def _fromdict(cls):
    """
    Add a classmethod to create an instance from
    a dictionary.
    """

    def _fromdict(cls, d):
        """Build an instance from a dictionary"""
        return cls(**d)

    cls.fromdict = classmethod(_fromdict)
    return cls


def attr_class_to_from_dict(cls):
    """
    When a class is built using attrs, this decorator
    adds two methods, asdict and fromdict.  The latter
    is a classmethod to create an instance.
    """

    def _asdict(self):
        """Return dict representation"""
        return attr.asdict(self)

    cls.asdict = _asdict
    return _fromdict(cls)


def attr_class_to_from_dict_no_recurse(cls):
    """
    When a class is built using attrs, this decorator
    adds two methods, asdict and fromdict.  The latter
    is a classmethod to create an instance.

    This version does not recurse into nested objects.
    """

    def _asdict(self):
        """Return dict representation"""
        return attr.asdict(self, recurse=False)

    cls.asdict = _asdict

    return _fromdict(cls)


def _add_getstate(cls):
    """
    Add __getstate__ via the __dict__
    """

    def getstate(self):
        return self.asdict()

    cls.__getstate__ = getstate

    return cls


def attr_class_pickle(cls):
    """
    Simplistic pickling based
    on the __dict__
    """

    def setstate(self, d):
        self.__dict__.update(d)

    cls.__setstate__ = setstate

    return _add_getstate(cls)


def attr_class_pickle_with_super(cls):
    """
    Add simplistic pickling to a class
    that inherits from another.

    This is tied to our coding stanard,
    where a Python class inheriting from
    a C++ class must use common kwargs
    for both.

    In some cases, the C++ class requires
    more nuanced initialization, so we
    cannot use this decorator.
    """

    def setstate(self, d):
        self.__dict__.update(d)
        self.__attrs_post_init__()

    cls.__setstate__ = setstate
    return _add_getstate(cls)


def attr_add_asblack(cls):
    """
    The default __repr__ from attrs isn't readable
    for complex classes.  This adds a method
    for pretty-printing out the class using black's
    formatting rules
    """

    def _asblack(self):
        """
        Return a string representation formatted with black
        """
        import black

        # The try/except is to handle black's changing API
        try:
            return black.format_str(str(self), mode=black.Mode())
        except AttributeError:
            return black.format_str(str(self), mode=black.FileMode())

    cls.asblack = _asblack

    return cls


def region_custom_repr(cls):
    """
    Custom repr to correctly display weight for region instances
    """

    def custom_repr(self):
        name = self.__class__.__name__
        d = {k: v for k, v in self.__dict__.items()}
        if d["coupled"] is True:
            d["weight"] = (d["end"] - d["beg"]) * d["weight"]
        d = str(d)
        d = d.replace("{", "")
        d = d.replace("}", "")
        d = d.replace("'", "")
        d = d.replace(": ", "=")
        repr = f"fwdpy11.{name}({d})"
        return repr

    cls.__repr__ = custom_repr

    return cls



#
# Copyright (C) 2017-2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
from typing import Dict, Iterable, Optional

import numpy as np

from .._fwdpy11 import HaploidGenome, Mutation


class PopulationMixin(object):
    @property
    def ancient_sample_genetic_values(self) -> np.ndarray:
        """
        Return the genetic values for ancient samples as a readonly 2d
        numpy.ndarray.

        Rows are individuals.  Columns are genetic values.

        ..  versionadded 0.3.0

        """
        return self._ancient_sample_genetic_values  # type: ignore

    @property
    def fixations(self) -> Iterable[Mutation]:
        return self._fixations  # type: ignore

    @property
    def fixation_times(self) -> Iterable[int]:
        return self._fixation_times  # type: ignore

    @property
    def generation(self) -> int:
        return self._generation  # type: ignore

    @property
    def _is_simulating(self) -> bool:
        return self.is_simulating  # type: ignore

    @property
    def genetic_values(self) -> np.ndarray:
        """
        Return the genetic values as a readonly 2d numpy.ndarray.

        Rows are individuals.  Columns are genetic values.

        ..  versionadded 0.3.0

        """
        return self._genetic_values  # type: ignore

    @property
    def haploid_genomes(self) -> Iterable[HaploidGenome]:
        return self._haploid_genomes  # type: ignore

    @property
    def mcounts(self) -> Iterable[int]:
        return self._mcounts  # type: ignore

    @property
    def mcounts_ancient_samples(self) -> Iterable[int]:
        return self._mcounts_ancient_samples  # type: ignore

    def mutation_indexes(self, pos) -> Optional[np.ndarray]:
        """
        Return indexes in :attr:`fwdpy11.DiploidPopulation.mutations`
        that are mutations at postion `pos`.
        """
        return self._mutation_indexes(pos)  # type: ignore

    @property
    def mut_lookup(self) -> Dict:
        return self._mut_lookup  # type: ignore

    @property
    def mutations(self) -> Iterable[Mutation]:
        return self._mutations  # type: ignore

    @property
    def mutations_ndarray(self) -> np.ndarray:
        """
        Return readonly numpy.ndarray of mutation data.
        The data are returned as a copy.

        The esizes and heffects fields are not part of the
        array, as their size is not constant and therefore
        they canot be part of a numpy "dtype".

        .. versionadded:: 0.4.0
        """
        return self._mutations_ndarray  # type: ignore

    @property
    def N(self):
        return self._N  # type: ignore

    def find_fixation_by_key(self, key, offset=0):
        """
        Find a fixation by key.

        :param key: A mutation key. See :func:`fwdpy11.Mutation.key`.
        :type key: tuple
        :param offset: Offset to start search in fixation container.
        :type offset: int

        :rtype: int

        :returns: Index of fixation if found, or -1 otherwise.

        .. versionadded:: 0.2.0

        """
        return self._find_fixation_by_key(key, offset)  # type: ignore

    def find_mutation_by_key(self, key, offset):
        """
        Find a mutation by key.

        :param key: A mutation key. See :func:`fwdpy11.Mutation.key`.
        :type key: tuple
        :param offset: Offset to start search in mutation container.
        :type offset: int

        :rtype: int

        :returns: Index of mutation if found, or -1 otherwise.

        .. versionadded:: 0.2.0

        """
        return self._find_mutation_by_key(key, offset)  # type: ignore



#
# Copyright (C) 2017-2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

from typing import Iterable, List, Optional, Union

import numpy as np

from .._fwdpy11 import MutationRecord, Site, ll_TreeIterator
from .table_collection import TableCollection


class TreeIterator(ll_TreeIterator):
    """
    Iterate over the marginal trees in a :class:`fwdpy11.TableCollection`

    Positional arguments for initialization:

    :param tables: A table collection
    :type tables: TableCollection
    :param samples: List of sample nodes
    :type samples: list or numpy.ndarray

    The following are keyword only:

    :type update_samples: If `True`, track list of sample nodes for each tree
    :type update_samples: TableCollection
    :param ancient_samples: List of ancient_samples nodes
    :type ancient_samples: list or numpy.ndarray
    :param begin: Start iteration at this position
    :type begin: float
    :param end: End iteration at this position
    :type end: float or None

    If `end` is `None`, then iteration proceeds over all trees with
    after positions :math:`\\geq \\mathrm{begin}`.

    If `ancient_samples` is not `None`, then
    :meth:`TreeIterator.preserved_leaf_counts` will return the number
    of ancient samples below a node.


    .. versionadded 0.3.0

    .. versionchanged:: 0.4.1

        Add begin, end options as floats for initializing

    """

    def __init__(
        self,
        tables: TableCollection,
        samples: Union[List[int], np.ndarray],
        *,
        update_samples: bool = False,
        ancient_samples: Union[List[int], np.ndarray] = None,
        begin: float = 0.0,
        end: Optional[float] = None
    ):
        _end = end
        if _end is None:
            _end = float(np.finfo(np.float64).max)
        if ancient_samples is None:
            super(TreeIterator, self).__init__(
                tables, samples, update_samples, begin, _end
            )
        else:
            super(TreeIterator, self).__init__(
                tables, samples, ancient_samples, update_samples, begin, _end
            )

    def parent(self, u: int) -> int:
        """
        :param u: A node id
        :type u: int

        :return: Parent id of node `u`
        :rtype: int
        """
        return self._parent(u)

    def leaf_counts(self, u: int) -> int:
        """
        :param u: A node id
        :type u: int

        :return: number of samples (leaves) below node `u`
        :rtype: int
        """
        return self._leaf_counts(u)

    def preserved_leaf_counts(self, u: int) -> int:
        """
        :param u: A node id
        :type u: int

        :return: number of ancient samples below node `u`
        :rtype: int
        """
        return self._preserved_leaf_counts(u)

    def left_sib(self, u: int) -> int:
        """
        :param u: A node id
        :type u: int

        :return: Left sib id of node `u`
        :rtype: int
        """
        return self._left_sib(u)

    def right_sib(self, u: int) -> int:
        """
        :param u: A node id
        :type u: int

        :return: Right sib id of node `u`
        :rtype: int
        """
        return self._right_sib(u)

    def left_child(self, u: int) -> int:
        """
        :param u: A node id
        :type u: int

        :return: Left child id of node `u`
        :rtype: int
        """
        return self._left_child(u)

    def right_child(self, u: int) -> int:
        """
        :param u: A node id
        :type u: int

        :return: Right child id of node `u`
        :rtype: int
        """
        return self._right_child(u)

    def total_time(self) -> float:
        """
        :return: Sum of branch lengths
        :rtype: float
        """
        return self._total_time(self.tables.nodes)

    def nodes(self) -> np.ndarray:
        """
        Return the nodes in the current tree.

        The return order is preorder.

        .. versionadded:: 0.5.1

        """
        return self._nodes()

    def samples(self) -> np.ndarray:
        """
        :return: The samples list
        :rtype: numpy.ndarray
        """
        return self._samples()

    def samples_below(self, node: int, sort=False) -> np.ndarray:
        """
        Return the list of samples descending from a node.

        :param node: A node id
        :type node: int
        :param sort: (False) Whether or not to sort sample node IDs.
        :type sort: bool

        .. note::

            Do not store these sample lists without making a "deep"
            copy.  The internal buffer is re-used.
        """
        return self._samples_below(node, sort)

    def sites(self) -> Iterable[Site]:
        """
        Return iterator over all :class:`fwdpy11.Site` objects
        on the current tree.

        .. versionadded:: 0.5.1

        """
        return self._sites()

    def mutations(self) -> Iterable[MutationRecord]:
        """
        Return iterator over all :class:`fwdpy11.MutationRecord` objects
        on the current tree.

        .. versionadded:: 0.5.1

        """
        return self._mutations()

    @property
    def left(self) -> float:
        """Left coordinate of current tree (inclusive)"""
        return self._left

    @property
    def right(self) -> float:
        """Right coordinate of current tree (exclusive)."""
        return self._right

    @property
    def roots(self) -> np.ndarray:
        """Roots of the current tree"""
        return self._roots

    @property
    def sample_size(self) -> int:
        return self._sample_size()

    @property
    def tables(self) -> TableCollection:
        """The underlying TableCollection"""
        return self._tables



#
# Copyright (C) 2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import demes
import fwdpy11
import pytest


@pytest.fixture
def pdict1():
    pd = {
        "nregions": [],
        "sregions": [fwdpy11.ExpS(0, 1, 1, -0.2)],
        "recregions": [fwdpy11.PoissonInterval(0, 1, 1e-3)],
        "gvalue": [fwdpy11.Additive(2.0)],
        "rates": (0, 1e-4, None),
        "simlen": 1000,
    }
    return pd


@pytest.fixture
def pdict2():
    pd = {
        "nregions": [],
        "sregions": [fwdpy11.ExpS(0, 1, 1, -0.2)],
        "recregions": [],
        "gvalue": [fwdpy11.Additive(2.0)],
        "rates": (0, 1e-4, None),
        "simlen": 1000,
    }
    return pd


@pytest.mark.parametrize("pop", [{"N": 100, "genome_length": 1}],
                         indirect=["pop"])
def test_single_model_params(pop, pdict1):
    demography = fwdpy11.ForwardDemesGraph.tubes([100], 1)
    pdict1["demography"] = demography
    mp = fwdpy11.ModelParams(**pdict1)

    ts = pop.dump_tables_to_tskit(model_params=mp)

    # reconstruct
    mp_rebuilt = fwdpy11.ModelParams(**eval(ts.metadata["model_params"]))

    assert mp == mp_rebuilt


@pytest.mark.parametrize("pop", [{"N": 100, "genome_length": 1}],
                         indirect=["pop"])
def test_multiple_model_params(pop, pdict1, pdict2):
    demography = fwdpy11.ForwardDemesGraph.tubes([100], 1)
    pdict1["demography"] = demography
    pdict2["demography"] = demography

    mp1 = fwdpy11.ModelParams(**pdict1)
    mp2 = fwdpy11.ModelParams(**pdict2)

    ts = pop.dump_tables_to_tskit(model_params={"phase1": mp1, "phase2": mp2})

    # reconstruct
    mp1_rebuilt = fwdpy11.ModelParams(
        **eval(ts.metadata["model_params"]["phase1"]))
    mp2_rebuilt = fwdpy11.ModelParams(
        **eval(ts.metadata["model_params"]["phase2"]))

    assert mp1 == mp1_rebuilt
    assert mp2 == mp2_rebuilt


@pytest.mark.parametrize("pop", [{"N": 100, "genome_length": 1}], indirect=["pop"])
def test_demes_graph(pop, gutenkunst):
    ts = pop.dump_tables_to_tskit(demes_graph=gutenkunst)

    # get the graph back out
    g = demes.Graph.fromdict(ts.metadata["demes_graph"])
    assert g == gutenkunst


@pytest.mark.parametrize("pop", [{"N": 100, "genome_length": 1}], indirect=["pop"])
def test_population_metadata(pop, gutenkunst):
    dm = fwdpy11.discrete_demography.from_demes(gutenkunst)
    demes_metadata = {}
    for key, value in dm.deme_labels.items():
        demes_metadata[key] = {"name": value}
    ts = pop.dump_tables_to_tskit(population_metadata=demes_metadata)

    for i in range(ts.tables.populations.num_rows):
        assert ts.population(i).metadata == demes_metadata[i]

    # Method 2: get the description out, too
    demes_metadata = {}
    for i, d in enumerate(gutenkunst.demes):
        demes_metadata[i] = {"name": d.name, "description": d.description}

    # Validate that dm thinks that things are in the same order
    for key, value in dm.deme_labels.items():
        assert demes_metadata[key]["name"] == value

    ts = pop.dump_tables_to_tskit(population_metadata=demes_metadata)
    for i in range(ts.tables.populations.num_rows):
        md = ts.population(i).metadata
        assert md["name"] == gutenkunst.demes[i].name
        assert md["description"] == gutenkunst.demes[i].description


@pytest.mark.parametrize("pop", [{"N": 100, "genome_length": 1}], indirect=["pop"])
def test_user_defined_data(pop):
    ts = pop.dump_tables_to_tskit(data={"mydata": 11})
    assert ts.metadata["data"]["mydata"] == 11

    class MyType(object):
        def __init__(self, x):
            self.x = x

        def __repr__(self):
            return f"MyType(x={self.x})"

    ts = pop.dump_tables_to_tskit(data=str(MyType(x=11)))
    assert eval(ts.metadata["data"]).x == 11


@pytest.mark.parametrize("pop", [{"N": 100, "genome_length": 1}], indirect=["pop"])
def test_seed(pop):
    ts = pop.dump_tables_to_tskit(seed=333)
    assert ts.metadata["seed"] == 333

    with pytest.raises(ValueError):
        _ = pop.dump_tables_to_tskit(seed=-333)



import unittest

import fwdpy11


class TestOldBinaryFileFormats(unittest.TestCase):
    def test_reading_045(self):
        pop = fwdpy11.DiploidPopulation.load_from_file("tests/v045.bin")
        assert isinstance(pop, fwdpy11.DiploidPopulation) is True
        for i in pop.tables.mutations:
            self.assertEqual(
                pop.mutations[i.key].pos, pop.tables.sites[i.site].position
            )


if __name__ == "__main__":
    unittest.main()



import unittest

from quick_pops import quick_nonneutral_slocus


class testWriteMutationLabel(unittest.TestCase):
    def test_write_label(self):
        pop = quick_nonneutral_slocus()
        for i in pop.mutations:
            if i.neutral is False:
                i.label = 3
            else:
                i.label = 17
        for i in pop.mutations:
            if i.neutral is False:
                self.assertEqual(i.label, 3)
            else:
                self.assertEqual(i.label, 17)


class testGetMutationKey(unittest.TestCase):
    def test_get_popgenmut_key(self):
        pop = quick_nonneutral_slocus()
        for i, j in zip(pop.mutations, pop.mcounts):
            if j > 0:
                self.assertEqual(i.key, (i.pos, i.s, i.g))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017-2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

from typing import List, Optional, Union

import numpy as np

from .._fwdpy11 import _data_matrix_from_tables, _make_data_matrix
from .._types import DataMatrix, DiploidPopulation, TableCollection


def make_data_matrix(
    pop: DiploidPopulation,
    samples: Union[List, np.ndarray],
    record_neutral: bool,
    record_selected: bool,
) -> DataMatrix:
    """
    Create a :class:`fwdpy11.DataMatrix` from a table collection.

    :param pop: A population
    :type pop: :class:`fwdpy11.PopulationBase`
    :param samples: A list of sample nodes
    :type samples: list
    :param record_neutral: If True, generate data for neutral variants
    :type record_neutral: bool
    :param record_selected: If True, generate data for selected variants
    :type record_selected: bool

    .. deprecated:: 0.3.0

       Prefer :func:`fwdpy11.data_matrix_from_tables`.

    """
    from warnings import warn

    warn(
        "fwdpy11.make_data_matrix is deprecated."
        " Please use fwdpy11.data_matrix_from_tables.",
        FutureWarning,
    )
    return DataMatrix(_make_data_matrix(pop, samples,
                                        record_neutral, record_selected))


def data_matrix_from_tables(
    tables: TableCollection,
    samples: Union[List, np.ndarray],
    *,
    record_neutral: bool = True,
    record_selected: bool = True,
    include_fixations: bool = False,
    begin: float = 0.0,
    end: Optional[float] = None,
) -> DataMatrix:
    """
    Create a :class:`fwdpy11.DataMatrix` from a table collection.

    :param tables: A TableCollection
    :type tables: fwdpy11.TableCollection
    :param samples: A list of sample nodes
    :type samples: list or :class:`numpy.ndarray`
    :param record_neutral: (True) If True, generate data for neutral variants
    :type record_neutral: bool
    :param record_selected: (True) If True, generate data for selected variants
    :type record_selected: bool
    :param include_selected: (False) Whether to include variants fixed in the sample
    :type include_selected: bool
    :param begin: (0.0) Start of range, inclusive
    :param end: (max float) End of range, exclusive

    :rtype: :class:`fwdpy11.DataMatrix`

    .. versionadded:: 0.3.0

    .. versionchanged:: 0.4.1

           Add begin, end options as floats

    .. versionchanged:: 0.5.0

           No longer requires :class:`fwdpy11.MutationVector` argument

    """

    if end is not None:
        _end = end
    else:
        _end = float(np.finfo(np.float64).max)
    return DataMatrix(
        _data_matrix_from_tables(
            tables,
            samples,
            record_neutral,
            record_selected,
            include_fixations,
            begin,
            _end,
        )
    )



#
# Copyright (C) 2017-2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest

import attr

import fwdpy11.class_decorators


@fwdpy11.class_decorators.attr_class_to_from_dict
@fwdpy11.class_decorators.attr_add_asblack
@attr.s(auto_attribs=True)
class FauxClass(object):
    a: float
    b: float


class TestClassDecorators(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.f = FauxClass(1.0, 3.0)

    def test_black_pretty_print(self):
        try:
            _ = self.f.asblack()
        except:  # NOQA
            self.fail("self.f.asblack() raised unexpected exception")

    def test_to_from_dict(self):
        d = self.f.asdict()
        self.assertEqual(d["a"], self.f.a)
        self.assertEqual(d["b"], self.f.b)
        f = FauxClass.fromdict(d)
        self.assertEqual(f.a, self.f.a)
        self.assertEqual(f.b, self.f.b)


if __name__ == "__main__":
    unittest.main()
